# Copyright 2015, 2018 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from nova.compute import task_states
from oslo_serialization import jsonutils
from pypowervm import const as pvm_const
from pypowervm.tests.tasks import util as tju
from pypowervm.tests import test_fixtures as pvm_fx
from pypowervm.tests.test_utils import pvmhttp
from pypowervm.wrappers import virtual_io_server as pvm_vios

from nova_powervm import conf as cfg
from nova_powervm.tests.virt.powervm.volume import test_driver as test_vol
from nova_powervm.virt.powervm import exception as exc
from nova_powervm.virt.powervm.volume import npiv

VIOS_FEED = 'fake_vios_feed2.txt'
VIOS_FEED_2 = 'fake_vios_feed.txt'

CONF = cfg.CONF


class TestNPIVAdapter(test_vol.TestVolumeAdapter):
    """Tests the NPIV Volume Connector Adapter."""

    def setUp(self):
        super(TestNPIVAdapter, self).setUp()

        self.adpt = self.useFixture(pvm_fx.AdapterFx()).adpt

        def resp(file_name):
            return pvmhttp.load_pvm_resp(
                file_name, adapter=self.adpt).get_response()
        self.vios_feed_resp = resp(VIOS_FEED)
        self.wwpn1 = '21000024FF649104'
        self.wwpn2 = '21000024FF649107'
        self.vios_uuid = '3443DB77-AED1-47ED-9AA5-3DB9C6CF7089'
        self.slot_mgr = mock.Mock()

        # Set up the transaction manager
        feed = pvm_vios.VIOS.wrap(self.vios_feed_resp)
        self.ft_fx = pvm_fx.FeedTaskFx(feed)
        self.useFixture(self.ft_fx)

        # Set up the mocks for the internal volume driver
        name = 'nova_powervm.virt.powervm.volume.npiv.NPIVVolumeAdapter.'
        self.mock_port_count_p = mock.patch(name + '_ports_per_fabric')
        self.mock_port_count = self.mock_port_count_p.start()
        self.mock_port_count.return_value = 1

        self.mock_fabric_names_p = mock.patch(name + '_fabric_names')
        self.mock_fabric_names = self.mock_fabric_names_p.start()
        self.mock_fabric_names.return_value = ['A']

        self.mock_fabric_ports_p = mock.patch(name + '_fabric_ports')
        self.mock_fabric_ports = self.mock_fabric_ports_p.start()
        self.mock_fabric_ports.return_value = [self.wwpn1, self.wwpn2]

        @mock.patch('pypowervm.wrappers.virtual_io_server.VIOS.getter')
        @mock.patch('nova_powervm.virt.powervm.vm.get_pvm_uuid')
        def init_vol_adpt(mock_pvm_uuid, mock_getter):
            con_info = {'serial': 'id',
                        'data': {'initiator_target_map': {'i1': ['t1'],
                                                          'i2': ['t2', 't3']},
                                 'target_lun': '1'}}
            mock_inst = mock.MagicMock()
            mock_pvm_uuid.return_value = '1234'

            # The getter can just return the VIOS values (to remove a read
            # that would otherwise need to be mocked).
            mock_getter.return_value = feed

            return npiv.NPIVVolumeAdapter(self.adpt, 'host_uuid', mock_inst,
                                          con_info)
        self.vol_drv = init_vol_adpt()

    def tearDown(self):
        super(TestNPIVAdapter, self).tearDown()

        self.mock_port_count_p.stop()
        self.mock_fabric_names_p.stop()
        self.mock_fabric_ports_p.stop()

    @mock.patch('pypowervm.tasks.vfc_mapper.add_map')
    def test_connect_volume(self, mock_add_map):
        # Mock
        self._basic_system_metadata(npiv.FS_UNMAPPED)
        self.slot_mgr.build_map.get_vfc_slots = mock.Mock(
            return_value=['62'])

        def add_map(vios_w, host_uuid, vm_uuid, port_map, **kwargs):
            self.assertIsInstance(vios_w, pvm_vios.VIOS)
            self.assertEqual('host_uuid', host_uuid)
            self.assertEqual('1234', vm_uuid)
            self.assertEqual(('21000024FF649104', 'AA BB'), port_map)
            return 'good'
        mock_add_map.side_effect = add_map

        # Test connect volume
        self.vol_drv.connect_volume(self.slot_mgr)

        # Verify that the appropriate connections were made.
        self.assertEqual(1, mock_add_map.call_count)
        mock_add_map.assert_called_once_with(
            mock.ANY, 'host_uuid', '1234', ('21000024FF649104', 'AA BB'),
            lpar_slot_num='62', provided={})
        self.assertEqual(1, self.ft_fx.patchers['update'].mock.call_count)
        self.assertEqual(npiv.FS_INST_MAPPED,
                         self.vol_drv._get_fabric_state('A'))

        # Verify the correct post execute methods were added to the feed task
        self.assertEqual('fab_slot_A_id',
                         self.vol_drv.stg_ftsk._post_exec[0].name)
        self.assertEqual('fab_A_id',
                         self.vol_drv.stg_ftsk._post_exec[1].name)

    def test_connect_volume_not_valid(self):
        """Validates that a connect will fail if in a bad state."""
        self.mock_inst_wrap.can_modify_io.return_value = False, 'Invalid I/O'
        self.assertRaises(exc.VolumeAttachFailed, self.vol_drv.connect_volume,
                          self.slot_mgr)

    @mock.patch('pypowervm.tasks.vfc_mapper.add_map')
    def test_connect_volume_inst_mapped(self, mock_add_map):
        """Test if already connected to an instance, don't do anything"""
        self._basic_system_metadata(npiv.FS_INST_MAPPED)
        mock_add_map.return_value = None
        self.slot_mgr.build_map.get_vfc_slots = mock.Mock(
            return_value=['62'])

        # Test subsequent connect volume calls when the fabric is mapped with
        # inst partition
        self.vol_drv.connect_volume(self.slot_mgr)

        # Verify
        self.assertEqual(1, mock_add_map.call_count)
        self.assertEqual(0, self.ft_fx.patchers['update'].mock.call_count)

        # Check the fabric state remains mapped to instance
        self.assertEqual(npiv.FS_INST_MAPPED,
                         self.vol_drv._get_fabric_state('A'))

    @mock.patch('pypowervm.tasks.vfc_mapper.find_vios_for_wwpn')
    @mock.patch('pypowervm.tasks.vfc_mapper.derive_npiv_map')
    @mock.patch('nova_powervm.virt.powervm.volume.npiv.NPIVVolumeAdapter.'
                '_set_fabric_meta')
    def test_ensure_phys_ports(self, set_fabric_meta, derive_npiv_map,
                               find_vios_for_wwpn):
        """Test that an npiv mapping gets rebuilt for stale wwpns"""

        self._basic_system_metadata(npiv.FS_INST_MAPPED, "stale-wwpn")
        port_maps = [('stale-wwpn', 'AA BB')]
        vios_wraps = [mock.MagicMock()]
        fab = 'A'

        # First the 'good' case where output should equal input
        port_maps = [(self.wwpn1, 'AA BB')]
        find_vios_for_wwpn.return_value = (vios_wraps[0], mock.MagicMock())
        maps = self.vol_drv._ensure_phys_ports_for_system(port_maps,
                                                          vios_wraps, fab)
        self.assertEqual(port_maps, maps)
        find_vios_for_wwpn.assert_called_once()
        derive_npiv_map.assert_not_called()

        # Now test the stale case - rebuild
        port_maps = [('stale-wwpn', 'AA BB')]
        expected_map = [(self.wwpn1, port_maps[0][1])]
        derive_npiv_map.return_value = expected_map
        find_vios_for_wwpn.return_value = (None, None)
        maps = self.vol_drv._ensure_phys_ports_for_system(port_maps,
                                                          vios_wraps, fab)
        self.assertEqual(expected_map, maps)
        derive_npiv_map.assert_called_once()

    def _basic_system_metadata(self, fabric_state, p_wwpn='21000024FF649104'):
        meta_fb_key = self.vol_drv._sys_meta_fabric_key('A')
        meta_fb_map = '%s,AA,BB' % p_wwpn
        meta_st_key = self.vol_drv._sys_fabric_state_key('A')
        self.vol_drv.instance.system_metadata = {meta_st_key: fabric_state,
                                                 meta_fb_key: meta_fb_map}

    def test_extend_volume(self):
        # Ensure the method is implemented
        self.vol_drv.extend_volume()

    @mock.patch('pypowervm.tasks.vfc_mapper.remove_maps')
    @mock.patch('pypowervm.tasks.vfc_mapper.find_vios_for_vfc_wwpns')
    def test_disconnect_volume(self, mock_find_vios, mock_remove_maps):
        # Mock Data
        self.vol_drv.instance.task_state = 'deleting'

        meta_key = self.vol_drv._sys_meta_fabric_key('A')
        meta_map = '21000024FF649104,AA,BB,21000024FF649105,CC,DD'
        self.vol_drv.instance.system_metadata = {meta_key: meta_map}
        mock_find_vios.return_value = (mock.Mock(uuid=self.vios_uuid),)

        # Invoke
        self.vol_drv.disconnect_volume(self.slot_mgr)

        # Two maps removed on one VIOS
        self.assertEqual(2, mock_remove_maps.call_count)
        self.assertEqual(1, self.ft_fx.patchers['update'].mock.call_count)

    @mock.patch('pypowervm.tasks.vfc_mapper.remove_maps')
    @mock.patch('nova_powervm.virt.powervm.volume.npiv.NPIVVolumeAdapter.'
                '_get_fabric_meta')
    def test_disconnect_volume_no_fabric_meta(self, mock_get_fabric_meta,
                                              mock_remove_maps):
        # Mock Data.  The fabric_names is set to A by setUp.
        # Force a None return
        self.vol_drv.instance.task_state = 'deleting'
        mock_get_fabric_meta.return_value = []

        # Invoke
        self.vol_drv.disconnect_volume(self.slot_mgr)

        # No mappings should have been removed
        self.assertFalse(mock_remove_maps.called)

    def test_disconnect_volume_not_valid(self):
        """Validates that a disconnect will fail if in a bad state."""
        self.mock_inst_wrap.can_modify_io.return_value = False, 'Bleh'
        self.assertRaises(exc.VolumeDetachFailed,
                          self.vol_drv.disconnect_volume, self.slot_mgr)

    @mock.patch('nova_powervm.virt.powervm.volume.npiv.NPIVVolumeAdapter.'
                '_fabric_names')
    def test_disconnect_volume_not_on_same_host(self, mock_names):
        """Validates a disconnect still removes mapping when host's differ."""
        self.vol_drv.instance.task_state = None
        self.vol_drv.instance.host = 'not_host_in_conf'

        self.vol_drv.disconnect_volume(self.slot_mgr)
        mock_names.assert_called_once()

    @mock.patch('nova_powervm.virt.powervm.volume.npiv.NPIVVolumeAdapter.'
                '_fabric_names')
    def test_disconnect_volume_spawning_not_on_same_host(self, mock_names):
        """Validates disconnect when instance is spawning on another host."""
        self.vol_drv.instance.task_state = 'spawning'
        self.vol_drv.instance.host = 'not_host_in_conf'

        self.vol_drv.disconnect_volume(self.slot_mgr)
        mock_names.assert_called_once()

    @mock.patch('nova_powervm.virt.powervm.volume.npiv.NPIVVolumeAdapter.'
                '_remove_maps_for_fabric')
    def test_disconnect_volume_no_op(self, mock_remove_maps):
        """Tests that when the task state is not set, connections are left."""
        # Invoke
        self.vol_drv.instance.task_state = None
        self.vol_drv.instance.host = None

        self.vol_drv.disconnect_volume(self.slot_mgr)

        # Verify
        self.assertEqual(0, mock_remove_maps.call_count)

    def test_disconnect_volume_no_op_other_state(self):
        """Tests that the deletion doesn't go through on certain states."""
        self.vol_drv.instance.task_state = task_states.RESUMING
        self.vol_drv.instance.host = CONF.host

        # Invoke
        self.vol_drv.disconnect_volume(self.slot_mgr)
        self.assertEqual(0, self.adpt.read.call_count)

    def test_connect_volume_no_map(self):
        """Tests that if the VFC Mapping exists, another is not added."""
        # Mock Data
        self.vol_drv._fabric_names.return_value = {}
        self.vol_drv.connection_info = {'data': {'initiator_target_map':
                                                 {'a': None, 'b': None},
                                                 'volume_id': 'vid'}}

        mock_mapping = mock.MagicMock()
        mock_mapping.client_adapter.wwpns = {'a', 'b'}

        mock_vios = mock.MagicMock()
        mock_vios.vfc_mappings = [mock_mapping]

        # Invoke
        self.vol_drv.connect_volume(self.slot_mgr)

    def test_min_xags(self):
        xags = self.vol_drv.min_xags()
        self.assertEqual(2, len(xags))
        self.assertIn(pvm_const.XAG.VIO_STOR, xags)
        self.assertIn(pvm_const.XAG.VIO_FMAP, xags)

    @mock.patch('nova_powervm.virt.powervm.volume.npiv.NPIVVolumeAdapter.'
                '_get_fabric_meta')
    def test_is_initial_wwpn(self, mock_fabric_meta):
        # The deleting state is for roll back on spawn.  Migrating is a
        # scenario where you can't be creating new wwpns
        mock_fabric_meta.return_value = [('21000024FF649104', 'virt1 virt2')]
        bad_states = [task_states.DELETING, task_states.MIGRATING]
        for state in bad_states:
            self.vol_drv.instance.task_state = state
            self.assertFalse(self.vol_drv._is_initial_wwpn(
                npiv.FS_UNMAPPED, 'a'))

        # Task state should still be bad.
        self.assertFalse(self.vol_drv._is_initial_wwpn(npiv.FS_UNMAPPED, 'a'))

        # Set a good task state, but fails due to the WWPNs already being
        # hosted
        self.vol_drv.instance.task_state = task_states.NETWORKING
        self.assertFalse(self.vol_drv._is_initial_wwpn(npiv.FS_UNMAPPED, 'a'))

        # Validate that having no fabric metadata returns that this is an
        # initial wwpn
        mock_fabric_meta.return_value = []
        self.assertTrue(self.vol_drv._is_initial_wwpn(npiv.FS_UNMAPPED, 'a'))

        # Validate that has fabric metadata of a different host, and therefore
        # is still a valid initial wwpn.  It is initial because it simulates
        # a reschedule on a new host.
        mock_fabric_meta.return_value = [('BAD_WWPN', 'virt1 virt2')]
        self.assertTrue(self.vol_drv._is_initial_wwpn(npiv.FS_UNMAPPED, 'a'))

        # And now no task state.
        self.vol_drv.instance.task_state = None
        self.assertTrue(self.vol_drv._is_initial_wwpn(npiv.FS_UNMAPPED, 'a'))

    def test_is_migration_wwpn(self):
        inst = self.vol_drv.instance

        # Migrating on different host
        inst.task_state = task_states.MIGRATING
        inst.host = 'Not Correct Host'
        self.assertTrue(self.vol_drv._is_migration_wwpn(npiv.FS_INST_MAPPED))

        # Try if the instance isn't mapped
        self.assertFalse(self.vol_drv._is_migration_wwpn(npiv.FS_UNMAPPED))

        # Simulate a rollback on the target host from a live migration failure
        inst.task_state = None
        self.assertTrue(self.vol_drv._is_migration_wwpn(npiv.FS_INST_MAPPED))

        # Mapped but on same host
        inst.task_state = task_states.MIGRATING
        inst.host = CONF.host
        self.assertFalse(self.vol_drv._is_migration_wwpn(npiv.FS_INST_MAPPED))

    @mock.patch('pypowervm.tasks.vfc_mapper.derive_npiv_map')
    def test_configure_wwpns_for_migration(self, mock_derive):
        # Mock out the fabric
        meta_fb_key = self.vol_drv._sys_meta_fabric_key('A')
        meta_fb_map = '21000024FF649104,AA,BB,21000024FF649105,CC,DD'
        self.vol_drv.instance.system_metadata = {meta_fb_key: meta_fb_map}

        # Mock out what the derive returns
        expected_map = [('21000024FF649104', 'BB AA'),
                        ('21000024FF649105', 'DD CC')]
        mock_derive.return_value = expected_map

        # Invoke
        resp_maps = self.vol_drv._configure_wwpns_for_migration('A')

        # Make sure the updated maps are returned
        expected = [('21000024FF649104', 'BB AA'),
                    ('21000024FF649105', 'DD CC')]
        self.assertEqual(expected, resp_maps)
        mock_derive.assert_called_with(
            mock.ANY, ['21000024FF649104', '21000024FF649107'],
            ['BB', 'AA', 'DD', 'CC'])

    @mock.patch('pypowervm.tasks.vfc_mapper.derive_npiv_map')
    def test_configure_wwpns_for_migration_existing(self, mock_derive):
        """Validates nothing is done if WWPNs are already flipped."""
        # Mock out the fabric
        meta_fb_key = self.vol_drv._sys_meta_fabric_key('A')
        meta_fb_map = '21000024FF649104,C05076079CFF0FA0,C05076079CFF0FA1'
        meta_fb_st_key = self.vol_drv._sys_fabric_state_key('A')
        meta_fb_st_val = npiv.FS_MIGRATING
        self.vol_drv.instance.system_metadata = {
            meta_fb_key: meta_fb_map, meta_fb_st_key: meta_fb_st_val}

        # Invoke
        resp_maps = self.vol_drv._configure_wwpns_for_migration('A')

        # Make sure that the order of the client WWPNs is not changed.
        expected = [('21000024FF649104', 'C05076079CFF0FA0 C05076079CFF0FA1')]
        self.assertEqual(expected, resp_maps)
        self.assertFalse(mock_derive.called)

    @mock.patch('pypowervm.tasks.vfc_mapper.build_wwpn_pair')
    @mock.patch('pypowervm.tasks.vfc_mapper.derive_npiv_map')
    def test_wwpns(self, mock_derive, mock_build_pair):
        """Tests that new WWPNs get generated properly."""
        # Mock Data
        mock_derive.return_value = [('21000024FF649104', 'AA BB'),
                                    ('21000024FF649105', 'CC DD')]
        self.adpt.read.return_value = self.vios_feed_resp

        meta_key = self.vol_drv._sys_meta_fabric_key('A')
        self.vol_drv.instance.system_metadata = {meta_key: None}

        # Invoke
        wwpns = self.vol_drv.wwpns()

        # Check
        self.assertListEqual(['AA', 'CC'], wwpns)
        self.assertEqual('21000024FF649104,AA,BB,21000024FF649105,CC,DD',
                         self.vol_drv.instance.system_metadata[meta_key])
        self.assertEqual(1, mock_derive.call_count)
        self.assertTrue(self.vol_drv.instance.save.called)

    @mock.patch('nova_powervm.virt.powervm.volume.npiv.NPIVVolumeAdapter.'
                '_get_fabric_state')
    def test_wwpns_on_sys_meta(self, mock_fabric_state):
        """Tests that previously stored WWPNs are returned."""
        # Mock
        mock_fabric_state.return_value = npiv.FS_INST_MAPPED
        self.vol_drv.instance.host = CONF.host
        self.vol_drv.instance.system_metadata = {
            self.vol_drv._sys_meta_fabric_key('A'): 'phys1,a,b,phys2,c,d'}

        # Invoke and Verify
        self.assertListEqual(['a', 'c'], self.vol_drv.wwpns())

    @mock.patch('nova_powervm.virt.powervm.volume.npiv.NPIVVolumeAdapter.'
                '_configure_wwpns_for_migration')
    @mock.patch('nova_powervm.virt.powervm.volume.npiv.NPIVVolumeAdapter.'
                '_is_migration_wwpn')
    @mock.patch('nova_powervm.virt.powervm.volume.npiv.NPIVVolumeAdapter.'
                '_is_initial_wwpn')
    @mock.patch('nova_powervm.virt.powervm.volume.npiv.NPIVVolumeAdapter.'
                '_get_fabric_state')
    def test_wwpns_for_migration(self, mock_fabric_state, mock_initial,
                                 mock_migration, mock_configure):
        """Tests that wwpns for migration are generated properly."""
        # Mock
        mock_fabric_state.return_value = npiv.FS_INST_MAPPED
        mock_initial.return_value = False
        mock_migration.return_value = True
        mock_configure.return_value = [('phys1', 'a b'), ('phys2', 'c d')]
        self.vol_drv.stg_ftsk = mock.MagicMock()

        # Invoke and Verify
        self.assertListEqual(['a', 'c'], self.vol_drv.wwpns())

        # Verify that on migration, the WWPNs are reversed.
        self.assertEqual(1, self.vol_drv.stg_ftsk.feed.reverse.call_count)

    @mock.patch('nova_powervm.virt.powervm.volume.npiv.NPIVVolumeAdapter.'
                '_get_fabric_state')
    def test_wwpns_bad_task_state(self, mock_fabric_state):
        """Tests behavior with a bad task state."""
        # Mock
        mock_fabric_state.return_value = npiv.FS_UNMAPPED
        self.vol_drv.instance.system_metadata = {
            self.vol_drv._sys_meta_fabric_key('A'): 'phys1,a,b,phys2,c,d'}

        # Invoke and Verify
        for state in [task_states.DELETING, task_states.MIGRATING]:
            self.vol_drv.instance.task_state = state
            self.assertListEqual(['a', 'c'], self.vol_drv.wwpns())

    @mock.patch('pypowervm.tasks.vfc_mapper.find_vios_for_vfc_wwpns')
    @mock.patch('nova_powervm.virt.powervm.volume.npiv.NPIVVolumeAdapter.'
                '_set_fabric_meta')
    @mock.patch('nova_powervm.virt.powervm.volume.npiv.NPIVVolumeAdapter.'
                '_get_fabric_meta')
    @mock.patch('nova_powervm.virt.powervm.volume.npiv.NPIVVolumeAdapter.'
                '_fabric_names')
    def test_post_live_migration_at_destination(
            self, mock_fabric_names, mock_get_fabric_meta,
            mock_set_fabric_meta, mock_find_wwpns):
        mock_fabric_names.return_value = ['A', 'B']
        mock_get_fabric_meta.side_effect = [
            [('S1', 'AA BB'), ('S2', 'CC DD')],
            [('S3', 'EE FF')]]

        # This represents the new physical WWPNs on the target server side.
        mock_find_wwpns.side_effect = [
            (None, mock.Mock(backing_port=mock.Mock(wwpn='T1'))),
            (None, mock.Mock(backing_port=mock.Mock(wwpn='T2'))),
            (None, mock.Mock(backing_port=mock.Mock(wwpn='T3')))]

        # Execute the test
        mig_vol_stor = {}
        self.vol_drv.post_live_migration_at_destination(mig_vol_stor)

        # Client WWPNs should be flipped and the new physical WWPNs should be
        # associated with them.
        mock_set_fabric_meta.assert_any_call(
            'A', [('T1', 'BB AA'), ('T2', 'DD CC')])
        mock_set_fabric_meta.assert_any_call(
            'B', [('T3', 'FF EE')])

        # Invoke a second time.  Should not 're-flip' or even call set.
        mock_set_fabric_meta.reset_mock()
        self.vol_drv.post_live_migration_at_destination(mig_vol_stor)
        self.assertFalse(mock_set_fabric_meta.called)

    @mock.patch('pypowervm.tasks.vfc_mapper.find_vios_for_vfc_wwpns')
    @mock.patch('nova_powervm.virt.powervm.volume.npiv.NPIVVolumeAdapter.'
                '_get_fabric_meta')
    @mock.patch('nova_powervm.virt.powervm.volume.npiv.NPIVVolumeAdapter.'
                '_fabric_names')
    def test_pre_live_migration_on_source(
            self, mock_fabric_names, mock_get_fabric_meta,
            mock_find_vios_for_vfc_wwpns):
        mock_fabric_names.return_value = ['A', 'B']
        mock_get_fabric_meta.side_effect = [
            [('11', 'AA BB'), ('22', 'CC DD')],
            [('33', 'EE FF')]]

        vios_wraps = pvm_vios.VIOS.wrap(tju.load_file(VIOS_FEED))
        vios1_w = vios_wraps[0]

        def mock_client_adpt(slot):
            return mock.Mock(client_adapter=mock.Mock(lpar_slot_num=slot))

        mock_find_vios_for_vfc_wwpns.side_effect = [
            (vios1_w, mock_client_adpt(1)), (vios1_w, mock_client_adpt(2)),
            (vios1_w, mock_client_adpt(3))]

        # Execute the test
        mig_data = {}
        self.vol_drv.pre_live_migration_on_source(mig_data)

        self.assertEqual('[1, 2]', mig_data.get('src_npiv_fabric_slots_A'))
        self.assertEqual('[3]', mig_data.get('src_npiv_fabric_slots_B'))
        vios_peer_slots = jsonutils.loads(mig_data.get('src_vios_peer_slots'))
        self.assertItemsEqual([[1, 2, 3]], vios_peer_slots)
        # Ensure only string data is placed in the dict.
        for key in mig_data:
            self.assertEqual(str, type(mig_data[key]))

    @mock.patch('pypowervm.tasks.vfc_mapper.find_vios_for_vfc_wwpns')
    @mock.patch('nova_powervm.virt.powervm.volume.npiv.NPIVVolumeAdapter.'
                '_get_fabric_meta')
    @mock.patch('nova_powervm.virt.powervm.volume.npiv.NPIVVolumeAdapter.'
                '_fabric_names')
    def test_pre_live_migration_on_source_dual_vios(
            self, mock_fabric_names, mock_get_fabric_meta,
            mock_find_vios_for_vfc_wwpns):
        mock_fabric_names.return_value = ['A', 'B']
        mock_get_fabric_meta.side_effect = [
            [('11', 'AA BB'), ('22', 'CC DD')],
            [('33', 'EE FF'), ('44', 'GG HH')]]

        vios_wraps = pvm_vios.VIOS.wrap(tju.load_file(VIOS_FEED_2))
        vios1_w = vios_wraps[0]
        vios2_w = vios_wraps[1]

        def mock_client_adpt(slot):
            return mock.Mock(client_adapter=mock.Mock(lpar_slot_num=slot))

        mock_find_vios_for_vfc_wwpns.side_effect = [
            (vios1_w, mock_client_adpt(1)), (vios1_w, mock_client_adpt(2)),
            (vios2_w, mock_client_adpt(3)), (vios2_w, mock_client_adpt(4))]

        # Execute the test
        mig_data = {}
        self.vol_drv.pre_live_migration_on_source(mig_data)

        self.assertEqual('[1, 2]', mig_data.get('src_npiv_fabric_slots_A'))
        self.assertEqual('[3, 4]', mig_data.get('src_npiv_fabric_slots_B'))
        vios_peer_slots = jsonutils.loads(mig_data.get('src_vios_peer_slots'))
        self.assertItemsEqual([[1, 2], [3, 4]], vios_peer_slots)
        # Ensure only string data is placed in the dict.
        for key in mig_data:
            self.assertEqual(str, type(mig_data[key]))

        # test unequal mapping across 2 VIOS
        mock_get_fabric_meta.side_effect = [
            [('11', 'AA BB'), ('22', 'CC DD')],
            [('33', 'EE FF')]]

        mock_find_vios_for_vfc_wwpns.side_effect = [
            (vios1_w, mock_client_adpt(1)), (vios2_w, mock_client_adpt(2)),
            (vios1_w, mock_client_adpt(3))]

        # Execute the test
        mig_data = {}
        self.vol_drv.pre_live_migration_on_source(mig_data)

        self.assertEqual('[1, 2]', mig_data.get('src_npiv_fabric_slots_A'))
        self.assertEqual('[3]', mig_data.get('src_npiv_fabric_slots_B'))
        vios_peer_slots = jsonutils.loads(mig_data.get('src_vios_peer_slots'))
        self.assertItemsEqual([[1, 3], [2]], vios_peer_slots)
        # Ensure only string data is placed in the dict.
        for key in mig_data:
            self.assertEqual(str, type(mig_data[key]))

    @mock.patch('pypowervm.tasks.vfc_mapper.'
                'build_migration_mappings_for_fabric')
    @mock.patch('nova_powervm.virt.powervm.volume.npiv.NPIVVolumeAdapter.'
                '_fabric_names')
    def test_pre_live_migration_on_destination_legacy(
            self, mock_fabric_names, mock_build_mig_map):
        mock_fabric_names.return_value = ['A', 'B']

        mig_data = {'src_npiv_fabric_slots_A': jsonutils.dumps([1, 2]),
                    'src_npiv_fabric_slots_B': jsonutils.dumps([3])}

        mock_build_mig_map.side_effect = [['a'], ['b']]
        self.vol_drv.stg_ftsk = mock.MagicMock()

        # Execute the test
        self.vol_drv.pre_live_migration_on_destination(mig_data)

        self.assertEqual('["a"]', mig_data.get('dest_npiv_fabric_mapping_A'))
        self.assertEqual('["b"]', mig_data.get('dest_npiv_fabric_mapping_B'))
        # Ensure only string data is placed in the dict.
        for key in mig_data:
            self.assertEqual(str, type(mig_data[key]))

        # Order of the mappings is not important.
        self.assertEqual(
            {'b', 'a'},
            set(jsonutils.loads(mig_data.get('vfc_lpm_mappings'))))

        # Verify that on migration, the WWPNs are reversed.
        self.assertEqual(2, self.vol_drv.stg_ftsk.feed.reverse.call_count)

    @mock.patch('pypowervm.tasks.vfc_mapper.build_migration_mappings')
    @mock.patch('nova_powervm.virt.powervm.volume.npiv.NPIVVolumeAdapter.'
                '_fabric_names')
    def test_pre_live_migration_on_destination(
            self, mock_fabric_names, mock_build_mig_map):
        mock_fabric_names.return_value = ['A', 'B']

        mig_data = {'src_npiv_fabric_slots_A': jsonutils.dumps([1, 2]),
                    'src_npiv_fabric_slots_B': jsonutils.dumps([3]),
                    'src_vios_peer_slots': jsonutils.dumps([[1, 2, 3]])}

        mock_build_mig_map.return_value = {'a', 'b'}
        self.vol_drv.stg_ftsk = mock.MagicMock()

        # Execute the test
        self.vol_drv.pre_live_migration_on_destination(mig_data)

        # Order of the mappings is not important.
        self.assertEqual(
            {'b', 'a'}, set(jsonutils.loads(mig_data.get('vfc_lpm_mappings'))))

    def test_set_fabric_meta(self):
        port_map = [('1', 'aa AA'), ('2', 'bb BB'),
                    ('3', 'cc CC'), ('4', 'dd DD'),
                    ('5', 'ee EE'), ('6', 'ff FF'),
                    ('7', 'gg GG'), ('8', 'hh HH'),
                    ('9', 'ii II'), ('10', 'jj JJ')]
        expected = {'npiv_adpt_wwpns_A':
                    '1,aa,AA,2,bb,BB,3,cc,CC,4,dd,DD',
                    'npiv_adpt_wwpns_A_2':
                    '5,ee,EE,6,ff,FF,7,gg,GG,8,hh,HH',
                    'npiv_adpt_wwpns_A_3':
                    '9,ii,II,10,jj,JJ'}
        self.vol_drv.instance.system_metadata = dict()
        self.vol_drv._set_fabric_meta('A', port_map)
        self.assertEqual(self.vol_drv.instance.system_metadata, expected)

        # Clear out the metadata and make sure it sticks.
        self.vol_drv._set_fabric_meta('A', [])
        self.assertEqual(self.vol_drv.instance.system_metadata, {})

    def test_get_fabric_meta(self):
        system_meta = {'npiv_adpt_wwpns_A':
                       '1,aa,AA,2,bb,BB,3,cc,CC,4,dd,DD',
                       'npiv_adpt_wwpns_A_2':
                       '5,ee,EE,6,ff,FF,7,gg,GG,8,hh,HH',
                       'npiv_adpt_wwpns_A_3':
                       '9,ii,II,10,jj,JJ'}
        expected = [('1', 'aa AA'), ('2', 'bb BB'),
                    ('3', 'cc CC'), ('4', 'dd DD'),
                    ('5', 'ee EE'), ('6', 'ff FF'),
                    ('7', 'gg GG'), ('8', 'hh HH'),
                    ('9', 'ii II'), ('10', 'jj JJ')]
        self.vol_drv.instance.system_metadata = system_meta
        fabric_meta = self.vol_drv._get_fabric_meta('A')
        self.assertEqual(fabric_meta, expected)

    def test_vol_type(self):
        self.assertEqual('npiv', self.vol_drv.vol_type())



# Copyright 2017 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import six
from taskflow import task

from nova_powervm import conf as cfg
from nova_powervm.virt.powervm import exception as p_exc
from nova_powervm.virt.powervm import vm
from nova_powervm.virt.powervm.volume import driver as v_driver
from oslo_log import log as logging
from pypowervm import const as pvm_const
from pypowervm.tasks import client_storage as pvm_c_stor
from pypowervm.tasks import hdisk
from pypowervm.tasks import partition
from pypowervm.tasks import scsi_mapper as tsk_map
from pypowervm.wrappers import storage as pvm_stg

CONF = cfg.CONF
LOG = logging.getLogger(__name__)


@six.add_metaclass(abc.ABCMeta)
class RBDVolumeAdapter(v_driver.PowerVMVolumeAdapter):
    """Base class for connecting ceph based Cinder Volumes to PowerVM VMs."""

    def __init__(self, adapter, host_uuid, instance, connection_info,
                 stg_ftsk=None):
        super(RBDVolumeAdapter, self).__init__(
            adapter, host_uuid, instance, connection_info, stg_ftsk=stg_ftsk)
        self._nl_vios_ids = None

    @classmethod
    def min_xags(cls):
        return [pvm_const.XAG.VIO_SMAP]

    @classmethod
    def vol_type(cls):
        """The type of volume supported by this type."""
        return 'rbd'

    @property
    def vios_uuids(self):
        """List the UUIDs of the Virtual I/O Servers hosting the storage."""
        # Get the hosting UUID
        if self._nl_vios_ids is None:
            nl_vios_wrap = partition.get_mgmt_partition(self.adapter)
            self._nl_vios_ids = [nl_vios_wrap.uuid]
        return self._nl_vios_ids

    def pre_live_migration_on_destination(self, mig_data):
        """Perform pre live migration steps for the volume on the target host.

        This method performs any pre live migration that is needed.

        This method will be called after the pre_live_migration_on_source
        method.  The data from the pre_live call will be passed in via the
        mig_data.  This method should put its output into the dest_mig_data.

        :param mig_data: Dict of migration data for the destination server.
                         If the volume connector needs to provide
                         information to the live_migration command, it
                         should be added to this dictionary.
        """
        for vios_uuid in self.vios_uuids:
            exists, name = self.is_volume_on_vios(vios_uuid)
            if exists and name is not None:
                return
        name = self.connection_info["data"]["name"]
        LOG.warning("RBD %s not found", name, instance=self.instance)
        raise p_exc.VolumePreMigrationFailed(
            volume_id=self.volume_id, instance_name=self.instance.name)

    def _connect_volume(self, slot_mgr):
        name = self.connection_info["data"]["name"]
        volid = self.connection_info["data"]["volume_id"]
        user = CONF.powervm.rbd_user
        rbd = pvm_stg.RBD.bld_ref(self.adapter, name, tag=volid, user=user)

        def add_func(vios_w):
            # If the vios doesn't match, just return
            if vios_w.uuid not in self.vios_uuids:
                return None

            LOG.info("Adding rbd disk connection to VIOS %(vios)s.",
                     {'vios': vios_w.name}, instance=self.instance)
            slot, lua = slot_mgr.build_map.get_vscsi_slot(vios_w, name)
            if slot_mgr.is_rebuild and not slot:
                LOG.debug('Detected a device with path %(path)s on VIOS '
                          '%(vios)s on the rebuild that did not exist on the '
                          'source. Ignoring.',
                          {'path': name, 'vios': vios_w.uuid},
                          instance=self.instance)
                return None

            mapping = tsk_map.build_vscsi_mapping(
                self.host_uuid, vios_w, self.vm_uuid, rbd, lpar_slot_num=slot,
                lua=lua)
            return tsk_map.add_map(vios_w, mapping)

        self.stg_ftsk.add_functor_subtask(add_func)

        # Run after all the deferred tasks the query to save the slots in the
        # slot map.
        def set_slot_info():
            vios_wraps = self.stg_ftsk.feed
            partition_id = vm.get_vm_id(self.adapter, self.vm_uuid)
            for vios_w in vios_wraps:
                scsi_map = pvm_c_stor.udid_to_scsi_mapping(
                    vios_w, name, partition_id)
                if not scsi_map:
                    continue
                slot_mgr.register_vscsi_mapping(scsi_map)

        self.stg_ftsk.add_post_execute(task.FunctorTask(
            set_slot_info, name='rbd_slot_%s' % name))

    def _disconnect_volume(self, slot_mgr):
        # Build the match function
        name = self.connection_info["data"]["name"]
        match_func = tsk_map.gen_match_func(pvm_stg.VDisk,
                                            names=[name])

        # Make sure the remove function will run within the transaction manager
        def rm_func(vios_w):
            # If the vios doesn't match, just return
            if vios_w.uuid not in self.vios_uuids:
                return None

            LOG.info("Disconnecting instance %(inst)s from storage "
                     "disks.", {'inst': self.instance.name},
                     instance=self.instance)
            removed_maps = tsk_map.remove_maps(vios_w, self.vm_uuid,
                                               match_func=match_func)
            for rm_map in removed_maps:
                slot_mgr.drop_vscsi_mapping(rm_map)
            return removed_maps

        self.stg_ftsk.add_functor_subtask(rm_func)
        # Find the disk directly.
        vios_w = self.stg_ftsk.wrapper_tasks[self.vios_uuids[0]].wrapper
        mappings = tsk_map.find_maps(vios_w.scsi_mappings,
                                     client_lpar_id=self.vm_uuid,
                                     match_func=match_func)

        return [x.backing_storage for x in mappings]

    def is_volume_on_vios(self, vio):
        """Returns whether or not the volume file is on a VIOS.

        This method is used during live-migration and rebuild to
        check if the volume is available on the target host.

        :param vio: The Virtual I/O Server wrapper or UDID string of the VIOS.
        :return: True if the file is on the VIOS.  False
                 otherwise.
        :return: The file path.
        """
        try:
            vio_uuid = vio.uuid
        except AttributeError:
            vio_uuid = vio
        if vio_uuid not in self.vios_uuids:
            return False, None
        name = self.connection_info["data"]["name"]
        exists = hdisk.rbd_exists(self.adapter, vio_uuid, name)
        return exists, name if exists else None



# Copyright 2015, 2018 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
from oslo_concurrency import lockutils
from oslo_log import log as logging
from taskflow import task

from nova.compute import task_states
from oslo_serialization import jsonutils
from pypowervm import const as pvm_const
from pypowervm.tasks import client_storage as pvm_c_stor
from pypowervm.tasks import vfc_mapper as pvm_vfcm

from nova_powervm import conf as cfg
from nova_powervm.conf import powervm as pvm_cfg
from nova_powervm.virt.powervm import exception as exc
from nova_powervm.virt.powervm.i18n import _
from nova_powervm.virt.powervm.volume import driver as v_driver


LOG = logging.getLogger(__name__)
CONF = cfg.CONF

WWPN_SYSTEM_METADATA_KEY = 'npiv_adpt_wwpns'
FABRIC_STATE_METADATA_KEY = 'fabric_state'
FS_UNMAPPED = 'unmapped'
FS_MIGRATING = 'migrating'
FS_INST_MAPPED = 'inst_mapped'
TASK_STATES_FOR_DISCONNECT = [task_states.DELETING, task_states.SPAWNING]


class NPIVVolumeAdapter(v_driver.FibreChannelVolumeAdapter):
    """The NPIV implementation of the Volume Adapter.

    NPIV stands for N_Port ID Virtualization.  It is a means of providing
    more efficient connections between virtual machines and Fibre Channel
    backed SAN fabrics.

    From a management level, the main difference is that the Virtual Machine
    will have its own WWPNs and own Virtual FC adapter.  The Virtual I/O
    Server only passes through communication directly to the VM itself.
    """

    @classmethod
    def min_xags(cls):
        """List of pypowervm XAGs needed to support this adapter."""
        # Storage are so physical FC ports are available
        # FC mapping is for the connections between VIOS and client VM
        return [pvm_const.XAG.VIO_FMAP, pvm_const.XAG.VIO_STOR]

    @classmethod
    def vol_type(cls):
        """The type of volume supported by this type."""
        return 'npiv'

    def _connect_volume(self, slot_mgr):
        """Connects the volume.

        :param slot_mgr: A NovaSlotManager.  Used to store/retrieve the client
                         slots used when a volume is attached to the VM
        """
        # Run the add for each fabric.
        for fabric in self._fabric_names():
            self._add_maps_for_fabric(fabric, slot_mgr)

    def extend_volume(self):
        # The compute node does not need to take any additional steps for the
        # client to see the extended volume.
        pass

    def _disconnect_volume(self, slot_mgr):
        """Disconnect the volume.

        :param slot_mgr: A NovaSlotManager.  Used to delete the client slots
                         used when a volume is detached from the VM
        """
        # We should only delete the NPIV mappings if we are running through a
        # VM deletion.  VM deletion occurs when the task state is deleting.
        # However, it can also occur during a 'roll-back' of the spawn.
        # Disconnect of the volumes will only be called during a roll back
        # of the spawn. We also want to check that the instance is on this
        # host. If it isn't then we can remove the mappings because this is
        # being called as the result of an evacuation clean up.
        if (self.instance.task_state not in TASK_STATES_FOR_DISCONNECT):
            if (self.instance.host in [None, CONF.host]):
                # NPIV should only remove the VFC mapping upon a destroy of
                # the VM
                return

        # Run the disconnect for each fabric
        for fabric in self._fabric_names():
            self._remove_maps_for_fabric(fabric)

    def pre_live_migration_on_source(self, mig_data):
        """Performs pre live migration steps for the volume on the source host.

        Certain volume connectors may need to pass data from the source host
        to the target.  This may be required to determine how volumes connect
        through the Virtual I/O Servers.

        This method gives the volume connector an opportunity to update the
        mig_data (a dictionary) with any data that is needed for the target
        host during the pre-live migration step.

        Since the source host has no native pre_live_migration step, this is
        invoked from check_can_live_migrate_source in the overall live
        migration flow.

        :param mig_data: A dictionary that the method can update to include
                         data needed by the pre_live_migration_at_destination
                         method.
        """
        fabrics = self._fabric_names()
        vios_wraps = self.stg_ftsk.feed
        # This mapping contains the client slots used on a given vios.
        # { vios_uuid: [slot_num, ...], vios2_uuid: [slot_num2,..] }
        slot_peer_dict = dict()
        for fabric in fabrics:
            npiv_port_maps = self._get_fabric_meta(fabric)
            if not npiv_port_maps:
                continue

            client_slots = []
            for port_map in npiv_port_maps:
                vios_w, vfc_map = pvm_vfcm.find_vios_for_vfc_wwpns(
                    vios_wraps, port_map[1].split())
                slot_num = vfc_map.client_adapter.lpar_slot_num
                vios_uuid = vios_w.partition_uuid
                if vios_uuid not in slot_peer_dict:
                    slot_peer_dict[vios_uuid] = []
                slot_peer_dict[vios_uuid].append(slot_num)
                client_slots.append(slot_num)

            # Set the client slots into the fabric data to pass to the
            # destination. Only strings can be stored.
            mig_data['src_npiv_fabric_slots_%s' % fabric] = (
                jsonutils.dumps(client_slots))
        # The target really doesn't care what the UUID is of the source VIOS
        # it is on a different server.  So let's strip that out and just
        # get the values.
        mig_data['src_vios_peer_slots'] = (
            jsonutils.dumps(list(slot_peer_dict.values())))

    def pre_live_migration_on_destination(self, mig_data):
        """Perform pre live migration steps for the volume on the target host.

        This method performs any pre live migration that is needed.

        Certain volume connectors may need to pass data from the source host
        to the target.  This may be required to determine how volumes connect
        through the Virtual I/O Servers.

        This method will be called after the pre_live_migration_on_source
        method.  The data from the pre_live call will be passed in via the
        mig_data.  This method should put its output into the dest_mig_data.

        :param mig_data: Dict of migration data for the destination server.
                         If the volume connector needs to provide
                         information to the live_migration command, it
                         should be added to this dictionary.
        """
        vios_wraps = self.stg_ftsk.feed
        if 'src_vios_peer_slots' in mig_data:
            self._pre_live_migration_on_dest_new(vios_wraps, mig_data)
        else:
            self._pre_live_migration_on_dest_legacy(vios_wraps, mig_data)

    def _pre_live_migration_on_dest_new(self, vios_wraps, mig_data):
        # Need to first derive the port mappings that can be passed back
        # to the source system for the live migration call.  This tells
        # the source system what 'vfc mappings' to pass in on the live
        # migration command.
        fabric_data = dict()
        for fabric in self._fabric_names():
            fab_slots = jsonutils.loads(
                mig_data['src_npiv_fabric_slots_%s' % fabric])
            ports = self._fabric_ports(fabric)
            fabric_data[fabric] = {'slots': fab_slots,
                                   'p_port_wwpns': ports}

        slot_peers = jsonutils.loads(
            mig_data['src_vios_peer_slots'])
        fabric_mapping = pvm_vfcm.build_migration_mappings(
            vios_wraps, fabric_data, slot_peers)
        mig_data['vfc_lpm_mappings'] = jsonutils.dumps(fabric_mapping)

    def _pre_live_migration_on_dest_legacy(self, vios_wraps, mig_data):
        # Used in case the source server is running an old nova-compute (ex.
        # Mitaka).  To be removed in Ocata or Pike.
        #
        # Need to first derive the port mappings that can be passed back
        # to the source system for the live migration call.  This tells
        # the source system what 'vfc mappings' to pass in on the live
        # migration command.
        for fabric in self._fabric_names():
            slots = jsonutils.loads(
                mig_data['src_npiv_fabric_slots_%s' % fabric])
            fabric_mapping = pvm_vfcm.build_migration_mappings_for_fabric(
                vios_wraps, self._fabric_ports(fabric), slots)
            mig_data['dest_npiv_fabric_mapping_%s' % fabric] = (
                jsonutils.dumps(fabric_mapping))
            # Reverse the vios wrapper so that the other fabric will get the
            # on the second vios.
            vios_wraps.reverse()

        # Collate all of the individual fabric mappings into a single element.
        full_map = []
        for key, value in mig_data.items():
            if key.startswith('dest_npiv_fabric_mapping_'):
                full_map.extend(jsonutils.loads(value))
        mig_data['vfc_lpm_mappings'] = jsonutils.dumps(full_map)

    def post_live_migration_at_destination(self, mig_vol_stor):
        """Perform post live migration steps for the volume on the target host.

        This method performs any post live migration that is needed.  Is not
        required to be implemented.

        :param mig_vol_stor: An unbounded dictionary that will be passed to
                             each volume adapter during the post live migration
                             call.  Adapters can store data in here that may
                             be used by subsequent volume adapters.
        """
        vios_wraps = self.stg_ftsk.feed

        # This method will run on the target host after the migration is
        # completed.  Right after this the instance.save is invoked from the
        # manager.  Given that, we need to update the order of the WWPNs.
        # The first WWPN is the one that is logged into the fabric and this
        # will now indicate that our WWPN is logged in.
        LOG.debug('Post live migrate volume store: %s', mig_vol_stor,
                  instance=self.instance)
        for fabric in self._fabric_names():
            # We check the mig_vol_stor to see if this fabric has already been
            # flipped.  If so, we can continue.
            fabric_key = '%s_flipped' % fabric
            if mig_vol_stor.get(fabric_key, False):
                continue

            # Must not be flipped, so execute the flip
            npiv_port_maps = self._get_fabric_meta(fabric)
            new_port_maps = []
            for port_map in npiv_port_maps:
                # Flip the WPWNs
                c_wwpns = port_map[1].split()
                c_wwpns.reverse()
                LOG.debug('Flipping WWPNs, ports: %s wwpns: %s',
                          port_map, c_wwpns, instance=self.instance)
                # Get the new physical WWPN.
                vfc_map = pvm_vfcm.find_vios_for_vfc_wwpns(vios_wraps,
                                                           c_wwpns)[1]
                p_wwpn = vfc_map.backing_port.wwpn

                # Build the new map.
                new_map = (p_wwpn, " ".join(c_wwpns))
                new_port_maps.append(new_map)
            self._set_fabric_meta(fabric, new_port_maps)
            self._set_fabric_state(fabric, FS_INST_MAPPED)

            # Store that this fabric is now flipped.
            mig_vol_stor[fabric_key] = True

    def _is_initial_wwpn(self, fc_state, fabric):
        """Determines if the invocation to wwpns is for a general method.

        A 'general' method would be a spawn (with a volume) or a volume attach
        or detach.

        :param fc_state: The state of the fabric.
        :param fabric: The name of the fabric.
        :return: True if the invocation appears to be for a spawn/volume
                 action. False otherwise.
        """
        # Easy fabric state check.  If its a state other than unmapped, it
        # can't be an initial WWPN
        if fc_state != FS_UNMAPPED:
            return False

        # Easy state check.  This is important in case of a rollback failure.
        # If it is deleting or migrating, it definitely is not an initial WWPN
        if self.instance.task_state in [task_states.DELETING,
                                        task_states.MIGRATING]:
            return False

        # Next, we have to check the fabric metadata.  Having metadata
        # indicates that we have been on at least a single host.  However,
        # a VM could be rescheduled.  In that case, the 'physical WWPNs' won't
        # match.  So if any of the physical WWPNs are not supported by this
        # host, we know that it is 'initial' for this host.
        port_maps = self._get_fabric_meta(fabric)
        if len(port_maps) > 0 and self._hosts_wwpn(port_maps):
            return False

        # At this point, it should be correct.
        LOG.info("Instance has not yet defined a WWPN on fabric %(fabric)s. "
                 "Appropriate WWPNs will be generated.",
                 {'fabric': fabric}, instance=self.instance)
        return True

    def _hosts_wwpn(self, port_maps):
        """Determines if this system hosts the port maps.

        Hosting the port map will be determined if one of the physical WWPNs
        is hosted by one of the VIOSes.

        :param port_maps: The list of port mappings for the given fabric.
        """
        vios_wraps = self.stg_ftsk.feed
        if port_maps:
            for port_map in port_maps:
                for vios_w in vios_wraps:
                    for pfc_port in vios_w.pfc_ports:
                        if pfc_port.wwpn == port_map[0]:
                            return True
        return False

    def _is_migration_wwpn(self, fc_state):
        """Determines if the WWPN call is occurring during a migration.

        This determines if it is on the target host.

        :param fc_state: The fabrics state.
        :return: True if the instance appears to be migrating to this host.
                 False otherwise.
        """
        return fc_state == FS_INST_MAPPED and self.instance.host != CONF.host

    def _configure_wwpns_for_migration(self, fabric):
        """Configures the WWPNs for a migration.

        During a NPIV migration, the WWPNs need to be flipped.  This is because
        the second WWPN is what will be logged in on the source system.  So by
        flipping them, we indicate that the 'second' wwpn is the new one to
        log in.

        Another way to think of it is, this code should always return the
        correct WWPNs for the system that the workload will be running on.

        This WWPNs invocation is done on the target server prior to the
        actual migration call.  It is used to build the volume connector.
        Therefore this code simply flips the ports around.

        :param fabric: The fabric to configure.
        :return: An updated port mapping.
        """
        if self._get_fabric_state(fabric) == FS_MIGRATING:
            # If the fabric is migrating, just return the existing port maps.
            # They've already been flipped.
            return self._get_fabric_meta(fabric)

        # When we migrate...flip the WWPNs around.  This is so the other
        # WWPN logs in on the target fabric.  If this code is hit, the flip
        # hasn't yet occurred (read as first volume on the instance).
        port_maps = self._get_fabric_meta(fabric)
        client_wwpns = []
        for port_map in port_maps:
            c_wwpns = port_map[1].split()
            c_wwpns.reverse()
            client_wwpns.extend(c_wwpns)

        # Now derive the mapping to the VIOS physical ports on this system
        # (the destination)
        port_mappings = pvm_vfcm.derive_npiv_map(
            self.stg_ftsk.feed, self._fabric_ports(fabric), client_wwpns)

        # This won't actually get saved by the process.  The instance save will
        # only occur after the 'post migration'.  But if there are multiple
        # volumes, their WWPNs calls will subsequently see the data saved
        # temporarily here, and therefore won't "double flip" the wwpns back
        # to the original.
        self._set_fabric_meta(fabric, port_mappings)
        self._set_fabric_state(fabric, FS_MIGRATING)

        # Return the mappings
        return port_mappings

    @lockutils.synchronized('npiv_wwpns')
    def wwpns(self):
        """Builds the WWPNs of the adapters that will connect the ports."""
        # Refresh the instance.  It could have been updated by a concurrent
        # call from another thread to get the wwpns.
        self.instance.refresh()
        vios_wraps = self.stg_ftsk.feed
        resp_wwpns = []

        # If this is the first time to query the WWPNs for the instance, we
        # need to generate a set of valid WWPNs.  Loop through the configured
        # FC fabrics and determine if these are new, part of a migration, or
        # were already configured.
        for fabric in self._fabric_names():
            fc_state = self._get_fabric_state(fabric)
            LOG.info("NPIV wwpns fabric state=%(st)s.",
                     {'st': fc_state}, instance=self.instance)

            if self._is_initial_wwpn(fc_state, fabric):
                # Get a set of WWPNs that are globally unique from the system.
                v_wwpns = pvm_vfcm.build_wwpn_pair(
                    self.adapter, self.host_uuid,
                    pair_count=self._ports_per_fabric())

                # Derive the virtual to physical port mapping
                port_maps = pvm_vfcm.derive_npiv_map(
                    vios_wraps, self._fabric_ports(fabric), v_wwpns)

                # the fabric is mapped to the physical port) and the fabric
                # state.
                self._set_fabric_meta(fabric, port_maps)
                self._set_fabric_state(fabric, FS_UNMAPPED)
                self.instance.save()
            elif self._is_migration_wwpn(fc_state):
                # The migration process requires the 'second' wwpn from the
                # fabric to be used.
                port_maps = self._configure_wwpns_for_migration(fabric)
            else:
                # This specific fabric had been previously set.  Just pull
                # from the meta (as it is likely already mapped to the
                # instance)
                port_maps = self._get_fabric_meta(fabric)

            # Every loop through, we reverse the vios wrappers.  This is
            # done so that if Fabric A only has 1 port, it goes on the
            # first VIOS.  Then Fabric B would put its port on a different
            # VIOS.  This servers as a form of multi pathing (so that your
            # paths are not restricted to a single VIOS).
            vios_wraps.reverse()

            # Port map is set by either conditional, but may be set to None.
            # If not None, then add the WWPNs to the response.
            if port_maps is not None:
                for mapping in port_maps:
                    # Only add the first WWPN.  That is the one that will be
                    # logged into the fabric.
                    resp_wwpns.append(mapping[1].split()[0])

        # The return object needs to be a list for the volume connector.
        return resp_wwpns

    def _ensure_phys_ports_for_system(self, npiv_port_maps, vios_wraps,
                                      fabric):
        """Ensures that the npiv_port_map is correct for the system.

        Rare scenarios can occur where the physical port on the NPIV port
        map does not match the actual port.  This is generally caused when the
        last volume is removed from the VM, the VM is migrated to another host,
        and then a new volume is attached.

        Stale metadata would be there (as it can't be cleaned out) on the
        attach.  This method clears that up.

        :param npiv_port_maps: The existing port maps.
        :param vios_wraps: The Virtual I/O Server wraps.
        :param fabric: The name of the fabric
        :return: The npiv_port_maps.  May be unchanged.
        """
        # Check that all physical ports in the mappings belong to 'this'
        # set of VIOSs.
        if all(pvm_vfcm.find_vios_for_wwpn(vios_wraps, pm[0])[0]
               for pm in npiv_port_maps):
            LOG.debug("All physical ports were found on the given Virtual I/O "
                      "Server(s).", instance=self.instance)
            return npiv_port_maps

        # If ANY of the VIOS ports were not there, rebuild the port maps
        LOG.debug("Rebuild existing_npiv_port_maps=%s. Reset fabric state.",
                  npiv_port_maps, instance=self.instance)
        v_wwpns = []
        for port_map in npiv_port_maps:
            v_wwpns.extend(port_map[1].split())
        self._set_fabric_state(fabric, FS_UNMAPPED)

        # Derive new maps and don't preserve existing maps
        npiv_port_maps = pvm_vfcm.derive_npiv_map(
            vios_wraps, self._fabric_ports(fabric), v_wwpns, preserve=False)
        LOG.debug("Rebuilt port maps: %s", npiv_port_maps,
                  instance=self.instance)
        self._set_fabric_meta(fabric, npiv_port_maps)
        LOG.warning("Had to update the system metadata for the WWPNs due to "
                    "incorrect physical WWPNs on fabric %(fabric)s",
                    {'fabric': fabric}, instance=self.instance)

        return npiv_port_maps

    def _add_maps_for_fabric(self, fabric, slot_mgr):
        """Adds the vFC storage mappings to the VM for a given fabric.

        :param fabric: The fabric to add the mappings to.
        :param slot_mgr: A NovaSlotManager.  Used to store/retrieve the client
                         slots used when a volume is attached to the VM
        """
        vios_wraps = self.stg_ftsk.feed
        # Ensure the physical ports in the metadata are not for a different
        # host (stale). If so, rebuild the maps with current info.
        npiv_port_maps = self._ensure_phys_ports_for_system(
            self._get_fabric_meta(fabric), vios_wraps, fabric)
        volume_id = self.connection_info['serial']

        # This loop adds the maps from the appropriate VIOS to the client VM
        slot_ids = copy.deepcopy(slot_mgr.build_map.get_vfc_slots(
            fabric, len(npiv_port_maps)))
        for npiv_port_map in npiv_port_maps:
            vios_w = pvm_vfcm.find_vios_for_port_map(vios_wraps, npiv_port_map)
            if vios_w is None:
                LOG.error("Mappings were not able to find a proper VIOS. "
                          "The port mappings were %s.", npiv_port_maps,
                          instance=self.instance)
                raise exc.VolumeAttachFailed(
                    volume_id=volume_id, instance_name=self.instance.name,
                    reason=_("Unable to find a Virtual I/O Server that "
                             "hosts the NPIV port map for the server."))
            ls = [LOG.info, "Adding NPIV mapping for instance %(inst)s "
                            "for Virtual I/O Server %(vios)s.",
                  {'inst': self.instance.name, 'vios': vios_w.name}]

            # Add the subtask to add the specific map.
            slot_num = slot_ids.pop()
            self.stg_ftsk.wrapper_tasks[vios_w.uuid].add_functor_subtask(
                pvm_vfcm.add_map, self.host_uuid, self.vm_uuid, npiv_port_map,
                lpar_slot_num=slot_num, logspec=ls)

        # Store the client slot number for the NPIV mapping (for rebuild
        # scenarios)
        def set_vol_meta():
            vios_wraps = self.stg_ftsk.feed
            port_maps = self._get_fabric_meta(fabric)
            for port_map in port_maps:
                # The port map is [ 'phys_wwpn', 'client_wwpn1 client_wwpn2' ]
                # We only need one of the two client wwpns.
                vios_w = pvm_vfcm.find_vios_for_port_map(vios_wraps, port_map)
                c_wwpns = port_map[1].split()
                vfc_mapping = pvm_c_stor.c_wwpn_to_vfc_mapping(vios_w,
                                                               c_wwpns[0])

                # If there is no mapping, then don't add it.  It means that
                # the client WWPN is hosted on a different VIOS.
                if vfc_mapping is None:
                    continue

                # However, by this point we know that it is hosted on this
                # VIOS.  So the vfc_mapping will have the client adapter
                slot_mgr.register_vfc_mapping(vfc_mapping, fabric)

        self.stg_ftsk.add_post_execute(task.FunctorTask(
            set_vol_meta, name='fab_slot_%s_%s' % (fabric, volume_id)))

        # After all the mappings, make sure the fabric state is updated.
        def set_state():
            self._set_fabric_state(fabric, FS_INST_MAPPED)
        self.stg_ftsk.add_post_execute(task.FunctorTask(
            set_state, name='fab_%s_%s' % (fabric, volume_id)))

    def _remove_maps_for_fabric(self, fabric):
        """Removes the vFC storage mappings from the VM for a given fabric.

        :param fabric: The fabric to remove the mappings from.
        """
        npiv_port_maps = self._get_fabric_meta(fabric)
        if not npiv_port_maps:
            # If no mappings exist, exit out of the method.
            return

        vios_wraps = self.stg_ftsk.feed

        for npiv_port_map in npiv_port_maps:
            ls = [LOG.info, "Removing a NPIV mapping for instance "
                            "%(inst)s for fabric %(fabric)s.",
                  {'inst': self.instance.name, 'fabric': fabric}]
            vios_w = pvm_vfcm.find_vios_for_port_map(vios_wraps, npiv_port_map)

            if vios_w is not None:
                # Add the subtask to remove the specific map
                task_wrapper = self.stg_ftsk.wrapper_tasks[vios_w.uuid]
                task_wrapper.add_functor_subtask(
                    pvm_vfcm.remove_maps, self.vm_uuid,
                    port_map=npiv_port_map, logspec=ls)
            else:
                LOG.warning("No storage connections found between the Virtual "
                            "I/O Servers and FC Fabric %(fabric)s.",
                            {'fabric': fabric}, instance=self.instance)

    def _set_fabric_state(self, fabric, state):
        """Sets the fabric state into the instance's system metadata.

        :param fabric: The name of the fabric
        :param state: state of the fabric which needs to be set

         Possible Valid States:
         FS_UNMAPPED: Initial state unmapped.
         FS_INST_MAPPED: Fabric is mapped with the nova instance.
        """
        meta_key = self._sys_fabric_state_key(fabric)
        LOG.info("Setting Fabric state=%(st)s.",
                 {'st': state}, instance=self.instance)
        self.instance.system_metadata[meta_key] = state

    def _get_fabric_state(self, fabric):
        """Gets the fabric state from the instance's system metadata.

        :param fabric: The name of the fabric
        :return: The state of the fabric which needs to be set

         Possible Valid States:
         FS_UNMAPPED: Initial state unmapped.
         FS_INST_MAPPED: Fabric is mapped with the nova instance.
        """
        meta_key = self._sys_fabric_state_key(fabric)
        if self.instance.system_metadata.get(meta_key) is None:
            self.instance.system_metadata[meta_key] = FS_UNMAPPED

        return self.instance.system_metadata[meta_key]

    def _sys_fabric_state_key(self, fabric):
        """Returns the nova system metadata key for a given fabric."""
        return FABRIC_STATE_METADATA_KEY + '_' + fabric

    def _set_fabric_meta(self, fabric, port_map):
        """Sets the port map into the instance's system metadata.

        The system metadata will store per-fabric port maps that link the
        physical ports to the virtual ports.  This is needed for the async
        nature between the wwpns call (get_volume_connector) and the
        connect_volume (spawn).

        :param fabric: The name of the fabric.
        :param port_map: The port map (as defined via the derive_npiv_map
                         pypowervm method).
        """

        # We will store the metadata in comma-separated strings with up to 4
        # three-token pairs. Each set of three comprises the Physical Port
        # WWPN followed by the two Virtual Port WWPNs:
        # Ex:
        # npiv_wwpn_adpt_A:
        #     "p_wwpn1,v_wwpn1,v_wwpn2,p_wwpn2,v_wwpn3,v_wwpn4,..."
        # npiv_wwpn_adpt_A_2:
        #     "p_wwpn5,v_wwpn9,vwwpn_10,p_wwpn6,..."

        meta_elems = []
        for p_wwpn, v_wwpn in port_map:
            meta_elems.append(p_wwpn)
            meta_elems.extend(v_wwpn.split())

        LOG.info("Fabric %(fabric)s wwpn metadata will be set to %(meta)s.",
                 {'fabric': fabric, 'meta': ",".join(meta_elems)},
                 instance=self.instance)

        # Clear out the original metadata.  We may be reducing the number of
        # keys (ex. reschedule) so we need to just delete what we had before
        # we add something new.
        meta_key_root = self._sys_meta_fabric_key(fabric)
        for key in tuple(self.instance.system_metadata.keys()):
            if key.startswith(meta_key_root):
                del self.instance.system_metadata[key]

        # Build up the mapping for the new keys.
        fabric_id_iter = 1
        meta_key = meta_key_root
        key_len = len(meta_key)

        for key in range(self._get_num_keys(port_map)):
            start_elem = 12 * (fabric_id_iter - 1)
            meta_value = ",".join(meta_elems[start_elem:start_elem + 12])
            self.instance.system_metadata[meta_key] = meta_value
            # If this is not the first time through, replace the end else cat
            if fabric_id_iter > 1:
                fabric_id_iter += 1
                meta_key = meta_key.replace(meta_key[key_len:],
                                            "_%s" % fabric_id_iter)
            else:
                fabric_id_iter += 1
                meta_key = meta_key + "_%s" % fabric_id_iter

    def _get_fabric_meta(self, fabric):
        """Gets the port map from the instance's system metadata.

        See _set_fabric_meta.

        :param fabric: The name of the fabric.
        :return: The port map (as defined via the derive_npiv_map pypowervm
                 method.
        """
        meta_key = self._sys_meta_fabric_key(fabric)

        if self.instance.system_metadata.get(meta_key) is None:
            # If no mappings exist, log a warning.
            LOG.warning("No NPIV mappings exist for instance on fabric "
                        "%(fabric)s.  May not have connected to the fabric "
                        "yet or fabric configuration was recently modified.",
                        {'fabric': fabric}, instance=self.instance)
            return []

        wwpns = self.instance.system_metadata[meta_key]
        key_len = len(meta_key)
        iterator = 2
        meta_key = meta_key + "_" + str(iterator)
        while self.instance.system_metadata.get(meta_key) is not None:
            meta_value = self.instance.system_metadata[meta_key]
            wwpns += "," + meta_value
            iterator += 1
            meta_key = meta_key.replace(meta_key[key_len:],
                                        "_" + str(iterator))

        wwpns = wwpns.split(",")

        # Rebuild the WWPNs into the natural structure.
        return [(p, ' '.join([v1, v2])) for p, v1, v2
                in zip(wwpns[::3], wwpns[1::3], wwpns[2::3])]

    def _sys_meta_fabric_key(self, fabric):
        """Returns the nova system metadata key for a given fabric."""
        return WWPN_SYSTEM_METADATA_KEY + '_' + fabric

    def _fabric_names(self):
        """Returns a list of the fabric names."""
        return pvm_cfg.NPIV_FABRIC_WWPNS.keys()

    def _fabric_ports(self, fabric_name):
        """Returns a list of WWPNs for the fabric's physical ports."""
        return pvm_cfg.NPIV_FABRIC_WWPNS[fabric_name]

    def _ports_per_fabric(self):
        """Returns the number of virtual ports to be used per fabric."""
        return CONF.powervm.ports_per_fabric

    def _get_num_keys(self, port_map):
        """Returns the number of keys we need to generate"""
        # Keys will have up to 4 mapping pairs so we determine based on that
        if len(port_map) % 4 > 0:
            return int(len(port_map) / 4 + 1)
        else:
            return int(len(port_map) / 4)



# Copyright 2014, 2017 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from eventlet import greenthread
from nova.compute import power_state
from nova.compute import task_states
from nova import context as ctx
from nova import exception
from nova.virt import event
from oslo_concurrency import lockutils
from oslo_log import log as logging
from pypowervm import adapter as pvm_apt
from pypowervm import util as pvm_util
from pypowervm.wrappers import event as pvm_evt

from nova_powervm.virt.powervm import vm


LOG = logging.getLogger(__name__)

_INST_ACTIONS_HANDLED = {'PartitionState', 'NVRAM'}
_NO_EVENT_TASK_STATES = {
    task_states.SPAWNING,
    task_states.RESIZE_MIGRATING,
    task_states.RESIZE_REVERTING,
    task_states.REBOOTING,
    task_states.REBOOTING_HARD,
    task_states.REBOOT_STARTED_HARD,
    task_states.PAUSING,
    task_states.UNPAUSING,
    task_states.SUSPENDING,
    task_states.RESUMING,
    task_states.POWERING_OFF,
    task_states.POWERING_ON,
    task_states.RESCUING,
    task_states.UNRESCUING,
    task_states.REBUILDING,
    task_states.REBUILD_SPAWNING,
    task_states.MIGRATING,
    task_states.DELETING,
    task_states.SOFT_DELETING,
    task_states.RESTORING,
    task_states.SHELVING,
    task_states.SHELVING_OFFLOADING,
    task_states.UNSHELVING,
}

_LIFECYCLE_EVT_LOCK = 'pvm_lifecycle_event'

_CONTEXT = None


def _get_instance(inst, pvm_uuid):
    global _CONTEXT
    if inst is not None:
        return inst
    with lockutils.lock('get_context_once'):
        if _CONTEXT is None:
            _CONTEXT = ctx.get_admin_context()
    LOG.debug('PowerVM Nova Event Handler: Getting inst for id %s', pvm_uuid)
    return vm.get_instance(_CONTEXT, pvm_uuid)


class PowerVMNovaEventHandler(pvm_apt.WrapperEventHandler):
    """Used to receive and handle events from PowerVM and convert to Nova."""
    def __init__(self, driver):
        self._driver = driver
        self._lifecycle_handler = PowerVMLifecycleEventHandler(self._driver)
        self._uuid_cache = {}

    def _get_inst_uuid(self, inst, pvm_uuid):
        """Retrieve instance UUID from cache keyed by the PVM UUID.

        :param inst: the instance object.
        :param pvm_uuid: the PowerVM uuid of the vm
        :return inst: the instance object.
        :return inst_uuid: The nova instance uuid
        """
        inst_uuid = self._uuid_cache.get(pvm_uuid)
        if not inst_uuid:
            inst = _get_instance(inst, pvm_uuid)
            inst_uuid = inst.uuid if inst else None
            if inst_uuid:
                self._uuid_cache[pvm_uuid] = inst_uuid
        return inst, inst_uuid

    def _handle_inst_event(self, inst, pvm_uuid, details):
        """Handle an instance event.

        This method will check if an instance event signals a change in the
        state of the instance as known to OpenStack and if so, trigger an
        event upward.

        :param inst: the instance object.
        :param pvm_uuid: the PowerVM uuid of the vm
        :param details: Parsed Details from the event
        :return inst: The nova instance, which may be None
        """
        # If the NVRAM has changed for this instance and a store is configured.
        if 'NVRAM' in details and self._driver.nvram_mgr is not None:
            # Schedule the NVRAM for the instance to be stored.
            # We'll need to fetch the instance object in the event we don't
            # have the object and the UUID isn't cached. By updating the
            # object reference here and returning it the process method will
            # save the object in its cache.
            inst, inst_uuid = self._get_inst_uuid(inst, pvm_uuid)
            if inst_uuid is None:
                return None

            LOG.debug('Handle NVRAM event for PowerVM LPAR %s', pvm_uuid)
            self._driver.nvram_mgr.store(inst_uuid)

        # If the state of the vm changed see if it should be handled
        if 'PartitionState' in details:
            self._lifecycle_handler.process(inst, pvm_uuid)

        return inst

    def process(self, events):
        """Process the event that comes back from PowerVM.

        :param events: The pypowervm Event wrapper.
        """
        inst_cache = {}
        for pvm_event in events:
            try:
                if pvm_event.etype in (pvm_evt.EventType.NEW_CLIENT,
                                       pvm_evt.EventType.CACHE_CLEARED):
                    # TODO(efried): Should we pull and check all the LPARs?
                    self._uuid_cache.clear()
                    continue
                # See if this uri (from data) ends with a PowerVM UUID.
                pvm_uuid = pvm_util.get_req_path_uuid(
                    pvm_event.data, preserve_case=True)
                if pvm_uuid is None:
                    continue
                # Is it an instance event?
                if not pvm_event.data.endswith('LogicalPartition/' + pvm_uuid):
                    continue

                # Are we deleting? Meaning we need to clear the cache entry.
                if pvm_event.etype == pvm_evt.EventType.DELETE_URI:
                    try:
                        del self._uuid_cache[pvm_uuid]
                    except KeyError:
                        pass
                    continue
                # Pull all the pieces of the event.
                details = (pvm_event.detail.split(',') if pvm_event.detail
                           else [])
                # Is it one we care about?
                if not _INST_ACTIONS_HANDLED & set(details):
                    continue

                inst_cache[pvm_event.data] = self._handle_inst_event(
                    inst_cache.get(pvm_event.data), pvm_uuid, details)

            except Exception:
                # We deliberately keep this exception clause as broad as
                # possible - we don't want *any* error to stop us from
                # attempting to process the next event.
                LOG.exception('Unable to process PowerVM event %s',
                              str(pvm_event))


class PowerVMLifecycleEventHandler(object):
    """Because lifecycle events are weird, we need our own handler.

    Lifecycle events that come back from the hypervisor are very 'surface
    value'.  They tell you that it started, stopped, migrated, etc...  However,
    multiple events may come in quickly that represent a bigger action.  For
    instance a restart will generate a stop and then a start rapidly.

    Nova being asynchronous can flip those events around.  Where the start
    would flow through before the stop.  That is bad.

    We need to make sure that these events that can be linked to bigger
    lifecycle events can be wiped out if the converse action is run against
    it.  Ex. Don't send a stop event up to nova if you received a start event
    shortly after it.
    """
    def __init__(self, driver):
        self._driver = driver
        self._delayed_event_threads = {}

    @lockutils.synchronized(_LIFECYCLE_EVT_LOCK)
    def _emit_event(self, pvm_uuid, inst):
        # Get the current state
        try:
            pvm_state = vm.get_vm_qp(self._driver.adapter, pvm_uuid,
                                     'PartitionState')
        except exception.InstanceNotFound:
            LOG.debug("LPAR %s was deleted while event was delayed.", pvm_uuid,
                      instance=inst)
            return

        LOG.debug('New state %s for partition %s', pvm_state, pvm_uuid,
                  instance=inst)

        inst = _get_instance(inst, pvm_uuid)
        if inst is None:
            LOG.debug("Not emitting LifecycleEvent: no instance for LPAR %s",
                      pvm_uuid)
            return

        # If we're in the middle of a nova-driven operation, no event necessary
        if inst.task_state in _NO_EVENT_TASK_STATES:
            LOG.debug("Not emitting LifecycleEvent: instance task_state is %s",
                      inst.task_state, instance=inst)
            return

        # See if it's really a change of state from what OpenStack knows
        transition = vm.translate_event(pvm_state, inst.power_state)
        if transition is None:
            LOG.debug("No LifecycleEvent necessary for pvm_state(%s) and "
                      "power_state(%s).", pvm_state,
                      power_state.STATE_MAP[inst.power_state], instance=inst)
            return

        # Log as if normal event
        lce = event.LifecycleEvent(inst.uuid, transition)
        LOG.info('Sending LifecycleEvent for instance state change to: %s',
                 pvm_state, instance=inst)
        self._driver.emit_event(lce)

        # Delete out the queue
        del self._delayed_event_threads[pvm_uuid]

    @lockutils.synchronized(_LIFECYCLE_EVT_LOCK)
    def process(self, inst, pvm_uuid):
        """Emits the event, or adds it to the queue for delayed emission.

        :param inst: The nova instance.  May be None.
        :param pvm_uuid: The PowerVM LPAR UUID.
        """
        # Cancel out the current delay event.  Can happen as it goes
        # from SHUTTING_DOWN to NOT_ACTIVATED, multiple delayed events
        # can come in at once.  Only want the last.
        if pvm_uuid in self._delayed_event_threads:
            self._delayed_event_threads[pvm_uuid].cancel()

        # Spawn in the background
        elem = greenthread.spawn_after(15, self._emit_event, pvm_uuid, inst)
        self._delayed_event_threads[pvm_uuid] = elem



# Copyright IBM Corp. and contributors
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_concurrency import lockutils
from oslo_log import log as logging

from nova_powervm import conf as cfg
from nova_powervm.virt.powervm import exception as p_exc
from nova_powervm.virt.powervm import vm
from nova_powervm.virt.powervm.volume import driver as v_driver
from nova_powervm.virt.powervm.volume import volume

from pypowervm import const as pvm_const
from pypowervm.tasks import hdisk
from pypowervm.tasks import partition as pvm_tpar
from pypowervm.utils import transaction as tx
from pypowervm.wrappers import virtual_io_server as pvm_vios

import base64
import six

CONF = cfg.CONF
LOG = logging.getLogger(__name__)

UDID_KEY = 'target_UDID'

# A global variable that will cache the physical WWPNs on the system.
_vscsi_pfc_wwpns = None


class PVVscsiFCVolumeAdapter(volume.VscsiVolumeAdapter,
                             v_driver.FibreChannelVolumeAdapter):
    """The vSCSI implementation of the Volume Adapter.

    For physical volumes, hosted to the VIOS through Fibre Channel, that
    connect to the VMs with vSCSI.

    vSCSI is the internal mechanism to link a given hdisk on the Virtual
    I/O Server to a Virtual Machine.  This volume driver will take the
    information from the driver and link it to a given virtual machine.
    """

    def __init__(self, adapter, host_uuid, instance, connection_info,
                 stg_ftsk=None):
        """Initializes the vSCSI Volume Adapter.

        :param adapter: The pypowervm adapter.
        :param host_uuid: The pypowervm UUID of the host.
        :param instance: The nova instance that the volume should connect to.
        :param connection_info: Comes from the BDM.
        :param stg_ftsk: (Optional) The pypowervm transaction FeedTask for the
                         I/O Operations.  If provided, the Virtual I/O Server
                         mapping updates will be added to the FeedTask.  This
                         defers the updates to some later point in time.  If
                         the FeedTask is not provided, the updates will be run
                         immediately when the respective method is executed.
        """
        super(PVVscsiFCVolumeAdapter, self).__init__(
            adapter, host_uuid, instance, connection_info, stg_ftsk=stg_ftsk)
        self._pfc_wwpns = None

    @classmethod
    def min_xags(cls):
        """List of pypowervm XAGs needed to support this adapter."""
        # SCSI mapping is for the connections between VIOS and client VM
        return [pvm_const.XAG.VIO_SMAP]

    @classmethod
    def vol_type(cls):
        """The type of volume supported by this type."""
        return 'vscsi'

    def pre_live_migration_on_destination(self, mig_data):
        """Perform pre live migration steps for the volume on the target host.

        This method performs any pre live migration that is needed.

        Certain volume connectors may need to pass data from the source host
        to the target.  This may be required to determine how volumes connect
        through the Virtual I/O Servers.

        This method will be called after the pre_live_migration_on_source
        method.  The data from the pre_live call will be passed in via the
        mig_data.  This method should put its output into the dest_mig_data.

        :param mig_data: Dict of migration data for the destination server.
                         If the volume connector needs to provide
                         information to the live_migration command, it
                         should be added to this dictionary.
        """
        volume_id = self.volume_id
        found = False

        # See the connect_volume for why this is a direct call instead of
        # using the tx_mgr.feed
        vios_wraps = pvm_vios.VIOS.get(self.adapter,
                                       xag=[pvm_const.XAG.VIO_STOR])

        # Iterate through host vios list to find valid hdisks.
        for vios_w in vios_wraps:
            status, device_name, udid = self._discover_volume_on_vios(
                vios_w, volume_id)
            # If we found one, no need to check the others.
            found = found or hdisk.good_discovery(status, device_name)
            # if valid udid is returned save in mig_data
            volume_key = 'vscsi-' + volume_id
            if udid is not None:
                mig_data[volume_key] = udid

        if not found or volume_key not in mig_data:
            ex_args = dict(volume_id=volume_id,
                           instance_name=self.instance.name)
            raise p_exc.VolumePreMigrationFailed(**ex_args)

    def post_live_migration_at_source(self, migrate_data):
        """Performs post live migration for the volume on the source host.

        This method can be used to handle any steps that need to taken on
        the source host after the VM is on the destination.

        :param migrate_data: volume migration data
        """
        # Get the udid of the volume to remove the hdisk for.  We can't
        # use the connection information because LPM 'refreshes' it, which
        # wipes out our data, so we use the data from the destination host
        # to avoid having to discover the hdisk to get the udid.
        udid = migrate_data.get('vscsi-' + self.volume_id)
        self._cleanup_volume(udid)

    def cleanup_volume_at_destination(self, migrate_data):
        """Performs volume cleanup after LPM failure on the dest host.

        This method can be used to handle any steps that need to taken on
        the destination host after the migration has failed.

        :param migrate_data: migration data
        """
        udid = migrate_data.get('vscsi-' + self.volume_id)
        self._cleanup_volume(udid)

    def is_volume_on_vios(self, vios_w):
        """Returns whether or not the volume is on a VIOS.

        :param vios_w: The Virtual I/O Server wrapper.
        :return: True if the volume driver's volume is on the VIOS.  False
                 otherwise.
        :return: The udid of the device.
        """
        status, device_name, udid = self._discover_volume_on_vios(
            vios_w, self.volume_id)
        return hdisk.good_discovery(status, device_name), udid

    def extend_volume(self):
        # The compute node does not need to take any additional steps for the
        # client to see the extended volume.
        pass

    def _discover_volume_on_vios(self, vios_w, volume_id):
        """Discovers an hdisk on a single vios for the volume.

        :param vios_w: VIOS wrapper to process
        :param volume_id: Volume to discover
        :returns: Status of the volume or None
        :returns: Device name or None
        :returns: UDID or None
        """
        # Get the initiatior WWPNs, targets and Lun for the given VIOS.
        vio_wwpns, t_wwpns, lun = self._get_hdisk_itls(vios_w)

        # Build the ITL map and discover the hdisks on the Virtual I/O
        # Server (if any).
        itls = hdisk.build_itls(vio_wwpns, t_wwpns, lun)
        if len(itls) == 0:
            LOG.debug('No ITLs for VIOS %(vios)s for volume %(volume_id)s.',
                      {'vios': vios_w.name, 'volume_id': volume_id},
                      instance=self.instance)
            return None, None, None

        device_id = self.connection_info.get('data', {}).get('pg83NAA')
        if device_id:
            device_id = base64.b64encode(device_id.encode())

        status, device_name, udid = hdisk.discover_hdisk(self.adapter,
                                                         vios_w.uuid, itls,
                                                         device_id=device_id)

        if hdisk.good_discovery(status, device_name):
            LOG.info('Discovered %(hdisk)s on vios %(vios)s for volume '
                     '%(volume_id)s. Status code: %(status)s.',
                     {'hdisk': device_name, 'vios': vios_w.name,
                      'volume_id': volume_id, 'status': status},
                     instance=self.instance)
        elif status == hdisk.LUAStatus.DEVICE_IN_USE:
            LOG.warning('Discovered device %(dev)s for volume %(volume)s '
                        'on %(vios)s is in use. Error code: %(status)s.',
                        {'dev': device_name, 'volume': volume_id,
                         'vios': vios_w.name, 'status': status},
                        instance=self.instance)

        return status, device_name, udid

    def _connect_volume_to_vio(self, vios_w, slot_mgr):
        """Attempts to connect a volume to a given VIO.

        :param vios_w: The Virtual I/O Server wrapper to connect to.
        :param slot_mgr: A NovaSlotManager.  Used to delete the client slots
                         used when a volume is detached from the VM

        :return: True if the volume was connected.  False if the volume was
                 not (could be the Virtual I/O Server does not have
                 connectivity to the hdisk).
        """
        status, device_name, udid = self._discover_volume_on_vios(
            vios_w, self.volume_id)

        # Get the slot and LUA to assign.
        slot, lua = slot_mgr.build_map.get_vscsi_slot(vios_w, udid)

        if slot_mgr.is_rebuild and not slot:
            LOG.debug('Detected a device with UDID %(udid)s on VIOS '
                      '%(vios)s on the rebuild that did not exist on the '
                      'source. Ignoring.', {'udid': udid, 'vios': vios_w.uuid},
                      instance=self.instance)
            return False

        if hdisk.good_discovery(status, device_name):
            volume_id = self.connection_info["data"]["volume_id"]
            # Found a hdisk on this Virtual I/O Server.  Add the action to
            # map it to the VM when the stg_ftsk is executed.
            with lockutils.lock(hash(self)):
                self._add_append_mapping(vios_w.uuid, device_name,
                                         lpar_slot_num=slot, lua=lua,
                                         tag=volume_id)

            # Save the UDID for the disk in the connection info.  It is
            # used for the detach.
            self._set_udid(udid)
            LOG.debug('Added deferred task to attach device %(device_name)s '
                      'to vios %(vios_name)s.',
                      {'device_name': device_name, 'vios_name': vios_w.name},
                      instance=self.instance)

            # Valid attachment
            return True

        return False

    def _disconnect_volume(self, slot_mgr):
        """Disconnect the volume.

        :param slot_mgr: A NovaSlotManager.  Used to delete the client slots
                         used when a volume is detached from the VM
        """
        def discon_vol_for_vio(vios_w):
            """Removes the volume from a specific Virtual I/O Server.

            :param vios_w: The VIOS wrapper.
            :return: True if a remove action was done against this VIOS.  False
                     otherwise.
            """
            LOG.debug("Disconnect volume %(vol)s from vios uuid %(uuid)s",
                      dict(vol=self.volume_id, uuid=vios_w.uuid),
                      instance=self.instance)
            device_name = None
            udid = self._get_udid()
            try:
                if udid:
                    # This will only work if vios_w has the Storage XAG.
                    device_name = vios_w.hdisk_from_uuid(udid)

                if not udid or not device_name:
                    # We lost our bdm data. We'll need to discover it.
                    status, device_name, udid = self._discover_volume_on_vios(
                        vios_w, self.volume_id)

                    # Check if the hdisk is in a bad state in the I/O Server.
                    # Subsequent scrub code on future deploys will clean it up.
                    if not hdisk.good_discovery(status, device_name):
                        LOG.warning(
                            "Disconnect Volume: The backing hdisk for volume "
                            "%(volume_id)s on Virtual I/O Server %(vios)s is "
                            "not in a valid state.  This may be the result of "
                            "an evacuate.",
                            {'volume_id': self.volume_id, 'vios': vios_w.name},
                            instance=self.instance)
                        return False

            except Exception:
                LOG.exception(
                    "Disconnect Volume: Failed to find disk on Virtual I/O "
                    "Server %(vios_name)s for volume %(volume_id)s. Volume "
                    "UDID: %(volume_uid)s.",
                    {'vios_name': vios_w.name, 'volume_id': self.volume_id,
                     'volume_uid': udid}, instance=self.instance)
                return False

            # We have found the device name
            LOG.info("Disconnect Volume: Discovered the device %(hdisk)s "
                     "on Virtual I/O Server %(vios_name)s for volume "
                     "%(volume_id)s.  Volume UDID: %(volume_uid)s.",
                     {'volume_uid': udid, 'volume_id': self.volume_id,
                      'vios_name': vios_w.name, 'hdisk': device_name},
                     instance=self.instance)

            # Add the action to remove the mapping when the stg_ftsk is run.
            partition_id = vm.get_vm_id(self.adapter, self.vm_uuid)

            with lockutils.lock(hash(self)):
                self._add_remove_mapping(partition_id, vios_w.uuid,
                                         device_name, slot_mgr)

                # Add a step to also remove the hdisk
                self._add_remove_hdisk(vios_w, device_name)

            # Found a valid element to remove
            return True

        try:
            # See logic in _connect_volume for why this new FeedTask is here.
            discon_ftsk = tx.FeedTask(
                'discon_volume_from_vio', pvm_vios.VIOS.getter(
                    self.adapter, xag=[pvm_const.XAG.VIO_STOR]))
            # Find hdisks to disconnect
            discon_ftsk.add_functor_subtask(
                discon_vol_for_vio, provides='vio_modified', flag_update=False)
            ret = discon_ftsk.execute()

            # Warn if no hdisks disconnected.
            if not any([result['vio_modified']
                        for result in ret['wrapper_task_rets'].values()]):
                LOG.warning("Disconnect Volume: Failed to disconnect the "
                            "volume %(volume_id)s on ANY of the Virtual "
                            "I/O Servers.", {'volume_id': self.volume_id},
                            instance=self.instance)

        except Exception as e:
            LOG.exception('PowerVM error detaching volume from virtual '
                          'machine.', instance=self.instance)
            ex_args = {'volume_id': self.volume_id, 'reason': six.text_type(e),
                       'instance_name': self.instance.name}
            raise p_exc.VolumeDetachFailed(**ex_args)

    @lockutils.synchronized('vscsi_wwpns')
    def wwpns(self):
        """Builds the WWPNs of the adapters that will connect the ports.

        :return: The list of WWPNs that need to be included in the zone set.
        """
        # Use a global variable so this is pulled once when the process starts.
        global _vscsi_pfc_wwpns
        if _vscsi_pfc_wwpns is None:
            _vscsi_pfc_wwpns = pvm_tpar.get_physical_wwpns(self.adapter)
        return _vscsi_pfc_wwpns

    def _get_hdisk_itls(self, vios_w):
        """Returns the mapped ITLs for the hdisk for the given VIOS.

        A PowerVM system may have multiple Virtual I/O Servers to virtualize
        the I/O to the virtual machines. Each Virtual I/O server may have their
        own set of initiator WWPNs, target WWPNs and Lun on which hdisk is
        mapped. It will determine and return the ITLs for the given VIOS.

        :param vios_w: A virtual I/O Server wrapper.
        :return: List of the i_wwpns that are part of the vios_w,
        :return: List of the t_wwpns that are part of the vios_w,
        :return: Target lun id of the hdisk for the vios_w.
        """
        it_map = self.connection_info['data']['initiator_target_map']
        i_wwpns = it_map.keys()

        active_wwpns = vios_w.get_active_pfc_wwpns()
        vio_wwpns = [x for x in i_wwpns if x in active_wwpns]

        t_wwpns = []
        for it_key in vio_wwpns:
            t_wwpns.extend(it_map[it_key])
        lun = self.connection_info['data']['target_lun']

        return vio_wwpns, t_wwpns, lun



# Copyright 2016, 2017 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import hashlib
import os
import retrying
import six
import tempfile
import types

from nova_powervm import conf as cfg
from nova_powervm.conf import powervm
from nova_powervm.virt.powervm.i18n import _
from nova_powervm.virt.powervm.nvram import api

from oslo_concurrency import lockutils
from oslo_log import log as logging
from oslo_utils import excutils
from oslo_utils import uuidutils
from swiftclient import exceptions as swft_exc
from swiftclient import service as swft_srv


LOG = logging.getLogger(__name__)
CONF = cfg.CONF


class SwiftNvramStore(api.NvramStore):

    def __init__(self):
        super(SwiftNvramStore, self).__init__()
        self.container = CONF.powervm.swift_container
        # Build the swift service options
        self.options = self._init_swift_options()
        self.swift_service = swft_srv.SwiftService(options=self.options)
        self._container_found = False

    @staticmethod
    def _init_swift_options():
        """Initialize all the options needed to communicate with Swift."""

        for opt in powervm.swift_opts:
            if opt.required and getattr(CONF.powervm, opt.name) is None:
                raise api.NVRAMConfigOptionNotSet(option=opt.name)

        options = {
            'auth_version': CONF.powervm.swift_auth_version,
            'os_username': CONF.powervm.swift_username,
            'os_user_domain_name': CONF.powervm.swift_user_domain_name,
            'os_password': CONF.powervm.swift_password,
            'os_project_name': CONF.powervm.swift_project_name,
            'os_project_domain_name': CONF.powervm.swift_project_domain_name,
            'os_auth_url': CONF.powervm.swift_auth_url,
            'os_cacert': CONF.powervm.swift_cacert,
            'os_endpoint_type': CONF.powervm.swift_endpoint_type,
        }

        return options

    def _run_operation(self, f, *args, **kwargs):
        """Convenience method to call the Swift client service."""

        # Get the function to call
        func = getattr(self.swift_service, f)
        try:
            result = func(*args, **kwargs)
            # For generators we have to copy the results because the
            # service is going out of scope.
            if isinstance(result, types.GeneratorType):
                results = []
                LOG.debug('SwiftOperation results:')
                for r in result:
                    results.append(copy.deepcopy(r))
                    LOG.debug(str(r))
                result = results
            else:
                LOG.debug('SwiftOperation result: %s', str(result))
            return result
        except swft_srv.SwiftError:
            with excutils.save_and_reraise_exception(logger=LOG):
                LOG.exception("Error running swift operation.")

    @classmethod
    def _get_name_from_listing(cls, results):
        names = []
        for result in results:
            if result['success']:
                for obj in result['listing']:
                    names.append(obj['name'])
        return names

    def _get_container_names(self):
        results = self._run_operation('list', options={'long': True})
        return self._get_name_from_listing(results)

    def _get_object_names(self, container, prefix=None):
        # If this is the first pass, the container may not exist yet.  Check
        # to make sure it does, otherwise the list of the object names will
        # fail.
        if not self._container_found:
            container_names = self._get_container_names()
            self._container_found = (container in container_names)

            # If the container was still not found, then just return an empty
            # list.  There are no objects.
            if not self._container_found:
                return []

        results = self._run_operation(
            'list', options={'long': True, 'prefix': prefix},
            container=container)
        return self._get_name_from_listing(results)

    def _store(self, inst_key, data, exists=None):
        """Store the NVRAM into the storage service.

        :param inst_key: The key by which to store the data in the repository.
        :param data: the NVRAM data base64 encoded string
        :param exists: (Optional, Default: None) If specified, tells the upload
                       whether or not the object exists.  Should be a boolean
                       or None.  If left as None, the method will look up
                       whether or not it exists.
        """

        # If the object doesn't exist, we tell it to 'leave_segments'.  This
        # prevents a lookup and saves the logs from an ERROR in the swift
        # client (that really isn't an error...sigh).  It should be empty
        # if not the first upload (which defaults to leave_segments=False)
        # so that it overrides the existing element on a subsequent upload.
        if exists is None:
            exists = self._exists(inst_key)
        options = dict(leave_segments=True) if not exists else None

        # The swift client already has a retry opertaion. The retry method
        # takes a 'reset' function as a parameter. This parameter is 'None'
        # for all operations except upload. For upload, it's set to a default
        # method that throws a ClientException if the object to upload doesn't
        # implement tell/see/reset. If the authentication error occurs during
        # upload, this ClientException is raised with no retry. For any other
        # operation, swift client will retry and succeed.
        @retrying.retry(retry_on_result=lambda result: result,
                        wait_fixed=250, stop_max_attempt_number=2)
        def _run_upload_operation():
            """Run the upload operation

            Attempts retry for a maximum number of two times. The upload
            operation will fail with ClientException, if there is an
            authentication error. The second attempt only happens if the
            first attempt failed with ClientException. A return value of
            True means we should retry, and False means no failure during
            upload, thus no retry is required.

            Raises RetryError if the upload failed during second attempt,
            as the number of attempts for retry is reached.

            """
            source = six.StringIO(data)
            obj = swft_srv.SwiftUploadObject(source, object_name=inst_key)

            results = self._run_operation('upload', self.container,
                                          [obj], options=options)
            for result in results:
                if not result['success']:
                    # TODO(arun-mani - Bug 1611011): Filed for updating swift
                    # client to return http status code in case of failure
                    if isinstance(result['error'], swft_exc.ClientException):
                        # If upload failed during nvram/slot_map update due to
                        # expired keystone token, retry swift-client operation
                        # to allow regeneration of token
                        LOG.warning('NVRAM upload failed due to invalid '
                                    'token. Retrying upload.')
                        return True
                    # The upload failed.
                    raise api.NVRAMUploadException(instance=inst_key,
                                                   reason=result)
            return False
        try:
            _run_upload_operation()
        except retrying.RetryError as re:
            # The upload failed.
            reason = (_('Unable to store NVRAM after %d attempts') %
                      re.last_attempt.attempt_number)
            raise api.NVRAMUploadException(instance=inst_key, reason=reason)

    @lockutils.synchronized('nvram')
    def store(self, instance, data, force=True):
        """Store the NVRAM into the storage service.

        :param instance: The nova instance object OR instance UUID.
        :param data: the NVRAM data base64 encoded string
        :param force: boolean whether an update should always be saved,
                      otherwise, check to see if it's changed.
        """
        inst_uuid = (instance if
                     uuidutils.is_uuid_like(instance) else instance.uuid)
        exists = self._exists(inst_uuid)
        if not force and exists:
            # See if the entry exists and has not changed.
            results = self._run_operation('stat', options={'long': True},
                                          container=self.container,
                                          objects=[inst_uuid])
            result = results[0]
            if result['success']:
                existing_hash = result['headers']['etag']
                if six.PY3:
                    data = data.encode('ascii')
                md5 = hashlib.md5(data).hexdigest()
                if existing_hash == md5:
                    LOG.info(('NVRAM has not changed for instance with '
                             'UUID %s.'), inst_uuid)
                    return

        self._store(inst_uuid, data, exists=exists)
        LOG.debug('NVRAM updated for instance with UUID %s', inst_uuid)

    def store_slot_map(self, inst_key, data):
        """Store the Slot Map to Swift.

        :param inst_key: The instance key to use for the storage operation.
        :param data: The data of the object to store.  This should be a string.
        """
        self._store(inst_key, data)

    def fetch_slot_map(self, inst_key):
        """Fetch the Slot Map object.

        :param inst_key: The instance key to use for the storage operation.
        :returns: The slot map (as a string)
        """
        return self._fetch(inst_key)[0]

    def fetch(self, instance):
        """Fetch the NVRAM from the storage service.

        :param instance: The nova instance object or instance UUID.
        :returns: the NVRAM data base64 encoded string
        """
        inst_uuid = (instance if
                     uuidutils.is_uuid_like(instance) else instance.uuid)
        data, result = self._fetch(inst_uuid)
        if not data:
            raise api.NVRAMDownloadException(instance=inst_uuid,
                                             reason=result)
        return data

    def _exists(self, object_key):
        # Search by prefix, but since this is just a prefix, we need to loop
        # and do a check to make sure it fully looks.
        obj_names = self._get_object_names(self.container, prefix=object_key)
        for obj in obj_names:
            if object_key == obj:
                return True
        return False

    def _fetch(self, object_key):
        # Check if the object exists.  If not, return a result accordingly.
        if not self._exists(object_key):
            return None, _('Object does not exist in Swift.')

        try:
            # Create a temp file for download into
            with tempfile.NamedTemporaryFile(delete=False) as f:
                options = {
                    'out_file': f.name
                }
            # The file is now created and closed for the swift client to use.
            results = self._run_operation(
                'download', container=self.container, objects=[object_key],
                options=options)
            for result in results:
                if result['success']:
                    with open(f.name, 'r') as f:
                        return f.read(), result
                else:
                    return None, result
        finally:
            try:
                os.remove(f.name)
            except Exception:
                LOG.warning('Could not remove temporary file: %s', f.name)

    def delete_slot_map(self, inst_key):
        """Delete the Slot Map from Swift.

        :param inst_key: The instance key to use for the storage operation.
        """
        for result in self._run_operation('delete', container=self.container,
                                          objects=[inst_key]):

            LOG.debug('Delete slot map result: %s', str(result))
            if not result['success']:
                raise api.NVRAMDeleteException(reason=result,
                                               instance=inst_key)

    def delete(self, instance):
        """Delete the NVRAM into the storage service.

        :param instance: The nova instance object OR instance UUID.
        """
        inst_uuid = (instance if
                     uuidutils.is_uuid_like(instance) else instance.uuid)
        for result in self._run_operation('delete', container=self.container,
                                          objects=[inst_uuid]):

            LOG.debug('Delete result for instance with UUID %(inst_uuid)s: '
                      '%(res)s', {'inst_uuid': inst_uuid, 'res': result})
            if not result['success']:
                raise api.NVRAMDeleteException(instance=inst_uuid,
                                               reason=result)



# Copyright 2017 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from nova_powervm.tests.virt.powervm.volume import test_driver as test_vol
from nova_powervm.virt.powervm import exception as p_exc
from nova_powervm.virt.powervm.volume import rbd as v_drv
from pypowervm import const as pvm_const
from pypowervm.tests import test_fixtures as pvm_fx
from pypowervm.wrappers import base_partition as pvm_bp
from pypowervm.wrappers import storage as pvm_stg
from pypowervm.wrappers import virtual_io_server as pvm_vios


class FakeRBDVolAdapter(v_drv.RBDVolumeAdapter):
    """Subclass for RBDVolumeAdapter, since it is abstract."""

    def __init__(self, adapter, host_uuid, instance, connection_info,
                 stg_ftsk=None):
        super(FakeRBDVolAdapter, self).__init__(
            adapter, host_uuid, instance, connection_info, stg_ftsk=stg_ftsk)


class TestRBDVolumeAdapter(test_vol.TestVolumeAdapter):
    """Tests the RBDVolumeAdapter.  NovaLink is a I/O host."""

    def setUp(self):
        super(TestRBDVolumeAdapter, self).setUp()

        # Needed for the volume adapter
        self.adpt = self.useFixture(pvm_fx.AdapterFx()).adpt
        mock_inst = mock.MagicMock(uuid='2BC123')

        self.vol_drv = FakeRBDVolAdapter(
            self.adpt, 'host_uuid', mock_inst,
            {'data': {'name': 'pool/image', 'volume_id': 'a_vol_id'},
             'serial': 'volid1'})

        self.fake_vios = pvm_vios.VIOS.bld(
            self.adpt, 'vios1',
            pvm_bp.PartitionMemoryConfiguration.bld(self.adpt, 1024),
            pvm_bp.PartitionMemoryConfiguration.bld(self.adpt, 0.1, 1))
        self.feed = [pvm_vios.VIOS.wrap(self.fake_vios.entry)]
        ftskfx = pvm_fx.FeedTaskFx(self.feed)
        self.useFixture(ftskfx)

    def test_min_xags(self):
        """Ensures xag's only returns SCSI Mappings."""
        self.assertEqual([pvm_const.XAG.VIO_SMAP], self.vol_drv.min_xags())

    @mock.patch('pypowervm.tasks.scsi_mapper.add_map', autospec=True)
    @mock.patch('pypowervm.tasks.scsi_mapper.build_vscsi_mapping',
                autospec=True)
    @mock.patch('pypowervm.entities.Entry.uuid',
                new_callable=mock.PropertyMock)
    @mock.patch('pypowervm.tasks.slot_map.SlotMapStore.register_vscsi_mapping',
                autospec=True)
    @mock.patch('pypowervm.tasks.client_storage.udid_to_scsi_mapping',
                autospec=True)
    @mock.patch('nova_powervm.virt.powervm.vm.get_vm_id')
    @mock.patch('pypowervm.tasks.partition.get_mgmt_partition', autospec=True)
    @mock.patch('pypowervm.wrappers.storage.RBD.bld_ref')
    def test_connect_volume(self, mock_rbd_bld_ref, mock_get_mgmt_partition,
                            mock_get_vm_id, mock_udid_to_map, mock_reg_map,
                            mock_get_vios_uuid, mock_build_map, mock_add_map):
        # Mockups
        mock_rbd = mock.Mock()
        mock_rbd_bld_ref.return_value = mock_rbd
        mock_slot_mgr = mock.MagicMock()
        mock_slot_mgr.build_map.get_vscsi_slot.return_value = 4, 'fake_path'

        mock_vios = mock.Mock(uuid='uuid1')
        mock_get_mgmt_partition.return_value = mock_vios
        mock_get_vios_uuid.return_value = 'uuid1'
        mock_get_vm_id.return_value = 'partition_id'

        mock_udid_to_map.return_value = mock.Mock()
        mock_add_map.return_value = None

        # Set user
        v_drv.CONF.powervm.rbd_user = 'tester'
        # Invoke
        self.vol_drv.connect_volume(mock_slot_mgr)
        # Validate
        mock_rbd_bld_ref.assert_called_once_with(
            self.adpt, 'pool/image', tag='a_vol_id', user='tester')
        self.assertEqual(1, mock_build_map.call_count)
        self.assertEqual(1, mock_udid_to_map.call_count)

    @mock.patch('pypowervm.tasks.scsi_mapper.build_vscsi_mapping',
                autospec=True)
    @mock.patch('pypowervm.entities.Entry.uuid',
                new_callable=mock.PropertyMock)
    @mock.patch('pypowervm.tasks.slot_map.SlotMapStore.register_vscsi_mapping',
                autospec=True)
    @mock.patch('pypowervm.tasks.client_storage.udid_to_scsi_mapping',
                autospec=True)
    @mock.patch('nova_powervm.virt.powervm.vm.get_vm_id')
    @mock.patch('pypowervm.tasks.partition.get_mgmt_partition', autospec=True)
    @mock.patch('pypowervm.wrappers.storage.RBD.bld_ref')
    def test_connect_volume_rebuild_no_slot(
            self, mock_rbd_bld_ref, mock_get_mgmt_partition, mock_get_vm_id,
            mock_udid_to_map, mock_reg_map, mock_get_vios_uuid,
            mock_build_map):
        # Mockups
        mock_rbd = mock.Mock()
        mock_rbd_bld_ref.return_value = mock_rbd
        mock_slot_mgr = mock.MagicMock()
        mock_slot_mgr.is_rebuild = True
        mock_slot_mgr.build_map.get_vscsi_slot.return_value = None, None

        mock_vios = mock.Mock(uuid='uuid1')
        mock_get_mgmt_partition.return_value = mock_vios
        mock_get_vios_uuid.return_value = 'uuid1'
        # Invoke
        self.vol_drv.connect_volume(mock_slot_mgr)

        # Validate
        mock_rbd_bld_ref.assert_called_once_with(
            self.adpt, 'pool/image', tag='a_vol_id', user='')
        self.assertEqual(0, mock_build_map.call_count)

    @mock.patch('pypowervm.entities.Entry.uuid',
                new_callable=mock.PropertyMock)
    @mock.patch('pypowervm.tasks.partition.get_mgmt_partition', autospec=True)
    @mock.patch('pypowervm.tasks.scsi_mapper.gen_match_func', autospec=True)
    @mock.patch('pypowervm.tasks.scsi_mapper.find_maps', autospec=True)
    def test_disconnect_volume(self, mock_find_maps, mock_gen_match_func,
                               mock_get_mgmt_partition, mock_entry_uuid):
        # Mockups
        mock_slot_mgr = mock.MagicMock()

        mock_vios = mock.Mock(uuid='uuid1')
        mock_get_mgmt_partition.return_value = mock_vios

        mock_match_func = mock.Mock()
        mock_gen_match_func.return_value = mock_match_func
        mock_entry_uuid.return_value = 'uuid1'
        # Invoke
        self.vol_drv._disconnect_volume(mock_slot_mgr)

        # Validate
        mock_gen_match_func.assert_called_once_with(
            pvm_stg.VDisk, names=['pool/image'])
        mock_find_maps.assert_called_once_with(
            mock.ANY,
            client_lpar_id='2BC123', match_func=mock_match_func)

    @mock.patch('nova_powervm.virt.powervm.volume.rbd.RBDVolumeAdapter.'
                'vios_uuids', new_callable=mock.PropertyMock)
    @mock.patch('nova_powervm.virt.powervm.volume.rbd.RBDVolumeAdapter.'
                'is_volume_on_vios')
    def test_pre_live_migration_on_destination(self, mock_on_vios, mock_uuids):
        mock_uuids.return_value = ['uuid1']
        mock_on_vios.return_value = (True, 'pool/image')
        self.vol_drv.pre_live_migration_on_destination(mock.ANY)
        mock_on_vios.return_value = (False, None)
        self.assertRaises(
            p_exc.VolumePreMigrationFailed,
            self.vol_drv.pre_live_migration_on_destination, mock.ANY)

    @mock.patch('nova_powervm.virt.powervm.volume.rbd.RBDVolumeAdapter.'
                'vios_uuids', new_callable=mock.PropertyMock)
    @mock.patch('pypowervm.tasks.hdisk.rbd_exists', autospec=True)
    def test_is_volume_on_vios(self, mock_exists, mock_vios_uuids):
        mock_exists.return_value = True
        mock_vios_uuids.return_value = ['uuid1']
        vol_found, vol_name = self.vol_drv.is_volume_on_vios(
            mock.Mock(uuid='uuid2'))
        self.assertFalse(vol_found)
        self.assertIsNone(vol_name)
        vol_found, vol_name = self.vol_drv.is_volume_on_vios(
            mock.Mock(uuid='uuid1'))
        self.assertTrue(vol_found)
        self.assertEqual('pool/image', vol_name)
        mock_exists.return_value = False
        vol_found, vol_name = self.vol_drv.is_volume_on_vios(
            mock.Mock(uuid='uuid1'))
        self.assertFalse(vol_found)
        self.assertIsNone(vol_name)



# Copyright 2015, 2017 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import six

from oslo_log import log as logging
from pypowervm import exceptions as pvm_exc
from pypowervm.tasks import partition as pvm_partition
from pypowervm.tasks import storage as tsk_stg
from pypowervm.utils import transaction as pvm_tx
from pypowervm.wrappers import virtual_io_server as pvm_vios

from nova_powervm.virt.powervm import exception as exc
from nova_powervm.virt.powervm import vm


LOG = logging.getLogger(__name__)
LOCAL_FEED_TASK = 'local_feed_task'


@six.add_metaclass(abc.ABCMeta)
class PowerVMVolumeAdapter(object):
    """The volume adapter connects a Cinder volume to a VM.

    The role of the volume driver is to perform the connection between the
    compute node and the backing physical fabric.

    This volume adapter is a generic adapter for all volume types to extend.

    This is built similarly to the LibvirtBaseVolumeDriver.
    """
    def __init__(self, adapter, host_uuid, instance, connection_info,
                 stg_ftsk=None):
        """Initialize the PowerVMVolumeAdapter

        :param adapter: The pypowervm adapter.
        :param host_uuid: The pypowervm UUID of the host.
        :param instance: The nova instance that the volume should connect to.
        :param connection_info: The volume connection info generated from the
                                BDM. Used to determine how to connect the
                                volume to the VM.
        :param stg_ftsk: (Optional) The pypowervm transaction FeedTask for the
                         I/O Operations.  If provided, the Virtual I/O Server
                         mapping updates will be added to the FeedTask.  This
                         defers the updates to some later point in time.  If
                         the FeedTask is not provided, the updates will be run
                         immediately when the respective method is executed.
        """
        self.adapter = adapter
        self.host_uuid = host_uuid
        self.instance = instance
        self.connection_info = connection_info
        self.vm_uuid = vm.get_pvm_uuid(instance)
        # Lazy-set this
        self._vm_id = None

        self.reset_stg_ftsk(stg_ftsk=stg_ftsk)

    @property
    def vm_id(self):
        """Return the short ID (not UUID) of the LPAR for our instance.

        This method is unavailable during a pre live migration call since
        there is no instance of the VM on the destination host at the time.
        """
        if self._vm_id is None:
            self._vm_id = vm.get_vm_id(self.adapter, self.vm_uuid)
        return self._vm_id

    @property
    def volume_id(self):
        """Method to return the volume id.

        Every driver must implement this method if the default impl will
        not work for their data.
        """
        return self.connection_info['serial']

    def reset_stg_ftsk(self, stg_ftsk=None):
        """Resets the pypowervm transaction FeedTask to a new value.

        The previous updates from the original FeedTask WILL NOT be migrated
        to this new FeedTask.

        :param stg_ftsk: (Optional) The pypowervm transaction FeedTask for the
                         I/O Operations.  If provided, the Virtual I/O Server
                         mapping updates will be added to the FeedTask.  This
                         defers the updates to some later point in time.  If
                         the FeedTask is not provided, the updates will be run
                         immediately when this method is executed.
        """
        if stg_ftsk is None:
            getter = pvm_vios.VIOS.getter(self.adapter, xag=self.min_xags())
            self.stg_ftsk = pvm_tx.FeedTask(LOCAL_FEED_TASK, getter)
        else:
            self.stg_ftsk = stg_ftsk

    @classmethod
    def min_xags(cls):
        """List of pypowervm XAGs needed to support this adapter."""
        raise NotImplementedError()

    @property
    def vios_uuids(self):
        """List the UUIDs of the Virtual I/O Servers hosting the storage."""
        vios_wraps = pvm_partition.get_active_vioses(self.adapter)
        return [wrap.uuid for wrap in vios_wraps]

    @classmethod
    def vol_type(cls):
        """The type of volume supported by this driver."""
        raise NotImplementedError()

    def is_volume_on_vios(self, vios_w):
        """Returns whether or not the volume is on VIOS.

        This method is used in the NovaSlotManager to build slot map.
        Needs to be implemented within the subclass to support rebuild.

        :param vios_w: The Virtual I/O Server wrapper
        :return: True if the volume is available on the VIOS.  False
                 otherwise.
        :return: The unique identification of the volume.
        """
        raise NotImplementedError()

    def pre_live_migration_on_destination(self, mig_data):
        """Perform pre live migration steps for the volume on the target host.

        This method performs any pre live migration that is needed.

        Certain volume connectors may need to pass data from the source host
        to the target.  This may be required to determine how volumes connect
        through the Virtual I/O Servers.

        This method will be called after the pre_live_migration_on_source
        method.  The data from the pre_live call will be passed in via the
        mig_data.  This method should put its output into the dest_mig_data.

        :param mig_data: Dict of migration data for the destination server.
                         If the volume connector needs to provide
                         information to the live_migration command, it
                         should be added to this dictionary.
        """
        raise NotImplementedError()

    def pre_live_migration_on_source(self, mig_data):
        """Performs pre live migration steps for the volume on the source host.

        Certain volume connectors may need to pass data from the source host
        to the target.  This may be required to determine how volumes connect
        through the Virtual I/O Servers.

        This method gives the volume connector an opportunity to update the
        mig_data (a dictionary) with any data that is needed for the target
        host during the pre-live migration step.

        Since the source host has no native pre_live_migration step, this is
        invoked from check_can_live_migrate_source in the overall live
        migration flow.

        :param mig_data: A dictionary that the method can update to include
                         data needed by the pre_live_migration_at_destination
                         method.
        """
        pass

    def post_live_migration_at_source(self, migrate_data):
        """Performs post live migration for the volume on the source host.

        This method can be used to handle any steps that need to taken on
        the source host after the VM is on the destination.

        :param migrate_data: volume migration data
        """
        pass

    def post_live_migration_at_destination(self, mig_vol_stor):
        """Perform post live migration steps for the volume on the target host.

        This method performs any post live migration that is needed.  Is not
        required to be implemented.

        :param mig_vol_stor: An unbounded dictionary that will be passed to
                             each volume adapter during the post live migration
                             call.  Adapters can store data in here that may
                             be used by subsequent volume adapters.
        """
        pass

    def cleanup_volume_at_destination(self, migrate_data):
        """Performs volume cleanup after LPM failure on the dest host.

        This method can be used to handle any steps that need to taken on
        the destination host after the migration has failed.

        :param migrate_data: migration data
        """
        pass

    def connect_volume(self, slot_mgr):
        """Connects the volume.

        :param slot_mgr: A NovaSlotManager.  Used to store/retrieve the client
                         slots used when a volume is attached to the VM
        """
        # Check if the VM is in a state where the attach is acceptable.
        lpar_w = vm.get_instance_wrapper(self.adapter, self.instance)
        capable, reason = lpar_w.can_modify_io()
        if not capable:
            raise exc.VolumeAttachFailed(
                volume_id=self.volume_id, instance_name=self.instance.name,
                reason=reason)

        # Run the connect
        self._connect_volume(slot_mgr)

        if self.stg_ftsk.name == LOCAL_FEED_TASK:
            self.stg_ftsk.execute()

    def extend_volume(self):
        raise NotImplementedError()

    def _extend_volume(self, udid):
        """Rescan virtual disk so client VM can see extended size"""
        resized = False
        error = False
        for vios_uuid in self.vios_uuids:
            try:
                LOG.debug("Rescanning volume %(vol)s for vios uuid %(uuid)s",
                          dict(vol=self.volume_id, uuid=vios_uuid),
                          instance=self.instance)
                tsk_stg.rescan_vstor(vios_uuid, udid, adapter=self.adapter)
                resized = True
            except pvm_exc.VstorNotFound:
                LOG.info("Failed to find volume %(vol)s for VIOS "
                         "UUID %(uuid)s during extend operation.",
                         {'vol': self.volume_id, 'uuid': vios_uuid},
                         instance=self.instance)
            except pvm_exc.JobRequestFailed as e:
                error = True
                LOG.error("Failed to rescan volume %(vol)s for VIOS "
                          "UUID %(uuid)s. %(reason)s",
                          {'vol': self.volume_id, 'uuid': vios_uuid,
                           'reason': six.text_type(e)}, instance=self.instance)
        if not resized or error:
            raise exc.VolumeExtendFailed(volume_id=self.volume_id,
                                         instance_name=self.instance.name)

    def disconnect_volume(self, slot_mgr):
        """Disconnect the volume.

        :param slot_mgr: A NovaSlotManager.  Used to store/retrieve the client
                         slots used when a volume is detached from the VM.
        """
        # Check if the VM is in a state where the detach is acceptable.
        lpar_w = vm.get_instance_wrapper(self.adapter, self.instance)
        capable, reason = lpar_w.can_modify_io()
        if not capable:
            raise exc.VolumeDetachFailed(
                volume_id=self.volume_id, instance_name=self.instance.name,
                reason=reason)

        # Run the disconnect
        self._disconnect_volume(slot_mgr)

        if self.stg_ftsk.name == LOCAL_FEED_TASK:
            self.stg_ftsk.execute()

    def _connect_volume(self, slot_mgr):
        """Connects the volume.

        This is the actual method to implement within the subclass.  Some
        transaction maintenance is done by the parent class.

        :param slot_mgr: A NovaSlotStore.  Used to store/retrieve the client
                         slots used when a volume is attached to the VM.
        """
        raise NotImplementedError()

    def _disconnect_volume(self, slot_mgr):
        """Disconnect the volume.

        This is the actual method to implement within the subclass.  Some
        transaction maintenance is done by the parent class.

        :param slot_mgr: A NovaSlotManager.  Used to delete the client slots
                         used when a volume is detached from the VM
        """
        raise NotImplementedError()


@six.add_metaclass(abc.ABCMeta)
class FibreChannelVolumeAdapter(PowerVMVolumeAdapter):
    """Defines a Fibre Channel specific volume adapter.

    Fibre Channel has a few additional attributes for the volume adapter.
    This class defines the additional attributes so that the multiple FC
    sub classes can support them.
    """

    def wwpns(self):
        """Builds the WWPNs of the adapters that will connect the ports.

        :return: The list of WWPNs that need to be included in the zone set.
        """
        raise NotImplementedError()



# Copyright 2015, 2018 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import collections
import copy
from oslo_concurrency import lockutils
from oslo_log import log as logging

from nova import exception as nova_exc
from nova_powervm import conf as cfg
from nova_powervm.virt.powervm import exception as p_exc
from nova_powervm.virt.powervm import vm
from nova_powervm.virt.powervm.volume import driver as v_driver
from nova_powervm.virt.powervm.volume import volume
from pypowervm import const as pvm_const
from pypowervm import exceptions as pvm_exc
from pypowervm.tasks import hdisk
from pypowervm.tasks import partition as pvm_partition
from pypowervm.utils import transaction as tx
from pypowervm.wrappers import virtual_io_server as pvm_vios

from taskflow import task

import six


LOG = logging.getLogger(__name__)
CONF = cfg.CONF
DEVNAME_KEY = 'target_devname'
_ISCSI_INITIATORS = collections.OrderedDict()


def get_iscsi_initiators(adapter, vios_ids=None):
    """Gets the VIOS iSCSI initiators.

     For the first time invocation of this method after process start up,
     it populates initiators data for VIOSes (if specified, otherwise it
     gets active VIOSes from the host) and stores in memory for futher
     lookup.

    :param adapter: The pypowervm adapter
    :param vios_ids: List of VIOS ids to get the initiators. If not
                     specified, a list of active VIOSes for the
                     host is fetched (but only for the first time)
                     through the pypowervm adapter.
    :return: A dict of the form
             {<vios_id>: <list of initiators>}
    """

    global _ISCSI_INITIATORS

    def discover_initiator(vios_id):

        # Get the VIOS id lock for initiator lookup
        @lockutils.synchronized('inititator-lookup-' + vios_id)
        def _discover_initiator():
            if vios_id in _ISCSI_INITIATORS and _ISCSI_INITIATORS[vios_id]:
                return
            else:
                try:
                    initiator = hdisk.discover_iscsi_initiator(
                        adapter, vios_id)
                    _ISCSI_INITIATORS[vios_id] = initiator
                except (pvm_exc.ISCSIDiscoveryFailed,
                        pvm_exc.JobRequestFailed) as e:
                    # TODO(chhagarw): handle differently based on
                    # error codes
                    LOG.error(e)

        _discover_initiator()

    if vios_ids is None and not _ISCSI_INITIATORS:
        vios_list = pvm_partition.get_active_vioses(adapter)
        vios_ids = [vios.uuid for vios in vios_list]

    for vios_id in vios_ids or []:
        discover_initiator(vios_id)

    LOG.debug("iSCSI initiator info: %s" % _ISCSI_INITIATORS)
    return _ISCSI_INITIATORS


class IscsiVolumeAdapter(volume.VscsiVolumeAdapter,
                         v_driver.PowerVMVolumeAdapter):
    """The iSCSI implementation of the Volume Adapter.

    This driver will connect a volume to a VM. First using iSCSI to connect the
    volume to the I/O Host (NovaLink partition). Then using the PowerVM vSCSI
    technology to host it to the VM itself.
    """
    def __init__(self, adapter, host_uuid, instance, connection_info,
                 stg_ftsk=None):
        super(IscsiVolumeAdapter, self).__init__(
            adapter, host_uuid, instance, connection_info, stg_ftsk=stg_ftsk)
        if connection_info['driver_volume_type'] == 'iser':
            self.iface_name = 'iser'
        else:
            self.iface_name = CONF.powervm.iscsi_iface

    @classmethod
    def vol_type(cls):
        """The type of volume supported by this type."""
        return 'iscsi'

    @classmethod
    def min_xags(cls):
        """List of pypowervm XAGs needed to support this adapter."""
        return [pvm_const.XAG.VIO_SMAP]

    def pre_live_migration_on_destination(self, mig_data):
        """Perform pre live migration steps for the volume on the target host.

        This method performs any pre live migration that is needed.

        Certain volume connectors may need to pass data from the source host
        to the target.  This may be required to determine how volumes connect
        through the Virtual I/O Servers.

        This method will be called after the pre_live_migration_on_source
        method.  The data from the pre_live call will be passed in via the
        mig_data.  This method should put its output into the dest_mig_data.

        :param mig_data: Dict of migration data for the destination server.
                         If the volume connector needs to provide
                         information to the live_migration command, it
                         should be added to this dictionary.
        """

        # See the connect_volume for why this is a direct call instead of
        # using the tx_mgr.feed
        vios_wraps = pvm_vios.VIOS.get(self.adapter,
                                       xag=[pvm_const.XAG.VIO_STOR])

        volume_key = 'vscsi-' + self.volume_id
        for vios_w in vios_wraps:
            if vios_w.uuid not in self.vios_uuids:
                continue
            # Discover the volume on all VIOS's to trigger the
            # device configuration, which in turn will discover the
            # LUN associated with the volume. This needs to be
            # attempted on all VIOS's to determine the VIOS's that
            # will be servicing IO for this volume.
            udid = self._discover_volume_on_vios(vios_w)[1]
            if udid:
                LOG.debug("Discovered volume udid %(udid)s on vios %(name)s",
                          dict(udid=udid, name=vios_w.name))
                mig_data[volume_key] = udid

        if volume_key not in mig_data:
            LOG.debug("Failed to discover the volume")
            ex_args = dict(volume_id=self.volume_id,
                           instance_name=self.instance.name)
            raise p_exc.VolumePreMigrationFailed(**ex_args)

    def post_live_migration_at_destination(self, mig_data):
        """This method will update the connection info with the volume udid."""

        volume_key = 'vscsi-' + self.volume_id
        if volume_key in mig_data:
            self._set_udid(mig_data[volume_key])

    def post_live_migration_at_source(self, migrate_data):
        """Performs post live migration for the volume on the source host.

        This method can be used to handle any steps that need to taken on
        the source host after the VM is on the destination.

        :param migrate_data: volume migration data
        """
        # Get the udid of the volume to remove the hdisk for.  We can't
        # use the connection information because LPM 'refreshes' it, which
        # wipes out our data, so we use the data from the destination host
        # to avoid having to discover the hdisk to get the udid.
        udid = migrate_data.get('vscsi-' + self.volume_id)
        self._cleanup_volume(udid)

    def is_volume_on_vios(self, vios_w):
        """Returns whether or not the volume is on a VIOS.

        :param vios_w: The Virtual I/O Server wrapper.
        :return: True if the volume driver's volume is on the VIOS.  False
                 otherwise.
        :return: The udid of the device.
        """
        if vios_w.uuid not in self.vios_uuids:
            return False, None
        device_name, udid = self._discover_volume_on_vios(vios_w)
        return (device_name and udid) is not None, udid

    def _is_multipath(self):
        return self.connection_info["connector"].get("multipath", False)

    def _get_iscsi_conn_props(self, vios_w, auth=False):
        """Returns the required iSCSI connection properties."""
        props = dict()
        try:
            data = self.connection_info['data']
            # For multipath target properties should exist
            if all([key in data for key in ('target_portals',
                                            'target_iqns',
                                            'target_luns')]):
                props['target_portals'] = data['target_portals']
                props['target_iqns'] = data['target_iqns']
                props['target_luns'] = data['target_luns']

            if auth and 'discovery_auth_method' in data:
                for s in ('method', 'username', 'password'):
                    k = 'discovery_auth_' + s
                    props[k] = data[k]

            props['target_portal'] = data['target_portal']
            props['target_iqn'] = data['target_iqn']
            props['target_lun'] = data['target_lun']

            # if auth_method is set look for username, password
            if auth and 'auth_method' in data:
                props['auth_method'] = data['auth_method']
                props['auth_username'] = data['auth_username']
                props['auth_password'] = data['auth_password']

            return props

        except (KeyError, ValueError):
            # Missing information in the connection info
            LOG.warning('Failed to retrieve iSCSI connection properties '
                        'for vios %(vios)s, connection_info=%(cinfo)s',
                        dict(vios=vios_w.uuid, cinfo=self.connection_info))
        return None

    def _discover_vol(self, vios_w, props):
        portal = props.get("target_portals", props.get("target_portal"))
        iqn = props.get("target_iqns", props.get("target_iqn"))
        lun = props.get("target_luns", props.get("target_lun"))
        auth = props.get("auth_method")
        user = props.get("auth_username")
        password = props.get("auth_password")
        discovery_auth = props.get("discovery_auth_method")
        discovery_username = props.get("discovery_auth_username")
        discovery_password = props.get("discovery_auth_password")
        try:
            return hdisk.discover_iscsi(
                self.adapter, portal, user, password, iqn, vios_w.uuid,
                lunid=lun, iface_name=self.iface_name, auth=auth,
                discovery_auth=discovery_auth,
                discovery_username=discovery_username,
                discovery_password=discovery_password,
                multipath=self._is_multipath())
        except (pvm_exc.ISCSIDiscoveryFailed, pvm_exc.JobRequestFailed) as e:
            msg_args = {'vios': vios_w.uuid, 'err': six.text_type(e)}
            LOG.warning("iSCSI discovery on VIOS %(vios)s failed with "
                        "error: %(err)s", msg_args)
            return None, None

    def _discover_volume_on_vios(self, vios_w):
        """Discovers an hdisk on a single vios for the volume.

        :param vios_w: VIOS wrapper to process
        :returns: Device name or None
        :returns: LUN or None
        """
        device_name = udid = None
        conn_props = self._get_iscsi_conn_props(vios_w, auth=True)
        if conn_props is None:
            return None, None

        # Check if multipath, we can directly pass the IQN list to
        # to the low level driver for volume discovery, else iterate
        # over the IQN and get the list for discovery.
        if self._is_multipath():
            device_name, udid = self._discover_vol(vios_w, conn_props)
        else:
            for props in self._iterate_all_targets(conn_props):
                device_name, udid = self._discover_vol(vios_w, props)
        return device_name, udid

    def _connect_volume_to_vio(self, vios_w, slot_mgr):
        """Attempts to connect a volume to a given VIO.

        :param vios_w: The Virtual I/O Server wrapper to connect to.
        :param slot_mgr: A NovaSlotManager.  Used to delete the client slots
                         used when a volume is detached from the VM

        :return: True if the volume was connected.  False if the volume was
                 not (could be the Virtual I/O Server does not have
                 connectivity to the hdisk).
        """
        # check if the vios uuid exist in the expected vios list
        if vios_w.uuid not in self.vios_uuids:
            LOG.debug("Skipping connect volume %(vol)s from "
                      "inactive vios uuid %(uuid)s.",
                      dict(vol=self.volume_id, uuid=vios_w.uuid))
            return False

        device_name, udid = self._discover_volume_on_vios(vios_w)
        if device_name is not None and udid is not None:
            slot, lua = slot_mgr.build_map.get_vscsi_slot(vios_w, device_name)
            volume_id = self.connection_info["data"]["volume_id"]
            # Found a hdisk on this Virtual I/O Server.  Add the action to
            # map it to the VM when the stg_ftsk is executed.
            with lockutils.lock(hash(self)):
                self._add_append_mapping(
                    vios_w.uuid, device_name, lpar_slot_num=slot, lua=lua,
                    udid=udid, tag=volume_id)

            # Save the udid  for the disk in the connection info.  It is
            # used for the detach.
            self._set_udid(udid)

            LOG.debug('Device attached: %s', device_name,
                      instance=self.instance)

            # Valid attachment
            return True

        return False

    def extend_volume(self):
        """Rescan virtual disk so client VM can see extended size."""
        udid = self._get_udid()
        if udid is None:
            raise nova_exc.InvalidBDM()
        self._extend_volume(udid)

    def _disconnect_volume(self, slot_mgr):
        """Disconnect the volume.

        This is the actual method to implement within the subclass.  Some
        transaction maintenance is done by the parent class.

        :param slot_mgr: A NovaSlotManager.  Used to delete the client slots
                         used when a volume is detached from the VM
        """

        def discon_vol_for_vio(vios_w):
            """Removes the volume from a specific Virtual I/O Server.

            :param vios_w: The VIOS wrapper.
            :return: True if a remove action was done against this VIOS.  False
                     otherwise.
            """
            # Check if the vios uuid exist in the list
            if vios_w.uuid not in self.vios_uuids:
                LOG.debug("Skipping disconnect of volume %(vol)s from "
                          "inactive vios uuid %(uuid)s.",
                          dict(vol=self.volume_id, uuid=vios_w.uuid))
                return False

            LOG.debug("Disconnect volume %(vol)s from vios uuid %(uuid)s",
                      dict(vol=self.volume_id, uuid=vios_w.uuid),
                      instance=self.instance)

            device_name = None
            try:
                udid = self._get_udid()
                if udid:
                    # Get the device name using UniqueDeviceID Identifier.
                    device_name = vios_w.hdisk_from_uuid(udid)

                if not udid or not device_name:
                    # If we have no device name, at this point
                    # we should not continue.  Subsequent scrub code on
                    # future deploys will clean this up.
                    LOG.warning(
                        "Disconnect Volume: The backing hdisk for volume "
                        "%(volume_id)s on Virtual I/O Server %(vios)s is "
                        "not in a valid state.  No disconnect "
                        "actions to be taken as volume is not healthy.",
                        {'volume_id': self.volume_id, 'vios': vios_w.name},
                        instance=self.instance)
                    return False

            except Exception:
                LOG.exception(
                    "Disconnect Volume: Failed to find device on Virtual I/O "
                    "Server %(vios_name)s for volume %(volume_id)s.",
                    {'vios_name': vios_w.name, 'volume_id': self.volume_id},
                    instance=self.instance)
                return False

            # We have found the device name
            LOG.info("Disconnect Volume: Discovered the device %(hdisk)s "
                     "on Virtual I/O Server %(vios_name)s for volume "
                     "%(volume_id)s.",
                     {'volume_id': self.volume_id,
                      'vios_name': vios_w.name, 'hdisk': device_name},
                     instance=self.instance)

            # Add the action to remove the mapping when the stg_ftsk is run.
            partition_id = vm.get_vm_id(self.adapter, self.vm_uuid)

            with lockutils.lock(hash(self)):
                self._add_remove_mapping(partition_id, vios_w.uuid,
                                         device_name, slot_mgr)
                conn_data = self._get_iscsi_conn_props(vios_w)
                if not conn_data:
                    return False
                iqn = conn_data.get("target_iqns", conn_data.get("target_iqn"))
                portal = conn_data.get("target_portals",
                                       conn_data.get("target_portal"))
                lun = conn_data.get("target_luns",
                                    conn_data.get("target_lun"))

                def remove():
                    try:
                        hdisk.remove_iscsi(
                            self.adapter, iqn, vios_w.uuid, lun=lun,
                            iface_name=self.iface_name, portal=portal,
                            multipath=self._is_multipath())
                    except (pvm_exc.ISCSIRemoveFailed,
                            pvm_exc.JobRequestFailed) as e:
                        LOG.warning(e)

                self.stg_ftsk.add_post_execute(task.FunctorTask(
                    remove, name='remove_%s_from_vios_%s' % (device_name,
                                                             vios_w.uuid)))

            # Found a valid element to remove
            return True

        try:
            # See logic in _connect_volume for why this new FeedTask is here.
            discon_ftsk = tx.FeedTask(
                'discon_volume_from_vio', pvm_vios.VIOS.getter(
                    self.adapter, xag=[pvm_const.XAG.VIO_STOR]))
            # Find hdisks to disconnect
            discon_ftsk.add_functor_subtask(
                discon_vol_for_vio, provides='vio_modified', flag_update=False)
            ret = discon_ftsk.execute()

            # Warn if no hdisks disconnected.
            if not any([result['vio_modified']
                        for result in ret['wrapper_task_rets'].values()]):
                LOG.warning(
                    "Disconnect Volume: Failed to disconnect the volume "
                    "%(volume_id)s on ANY of the Virtual I/O Servers.",
                    {'volume_id': self.volume_id}, instance=self.instance)

        except Exception as e:
            LOG.exception('PowerVM error detaching volume from virtual '
                          'machine.', instance=self.instance)
            ex_args = {'volume_id': self.volume_id, 'reason': six.text_type(e),
                       'instance_name': self.instance.name}
            raise p_exc.VolumeDetachFailed(**ex_args)

    # Taken from os_brick.initiator.connectors.base_iscsi.py
    def _iterate_all_targets(self, connection_properties):
        for portal, iqn, lun in self._get_all_targets(connection_properties):
            props = copy.deepcopy(connection_properties)
            props['target_portal'] = portal
            props['target_iqn'] = iqn
            props['target_lun'] = lun
            for key in ('target_portals', 'target_iqns', 'target_luns'):
                props.pop(key, None)
            yield props

    def _get_all_targets(self, connection_properties):
        if all([key in connection_properties for key in ('target_portals',
                                                         'target_iqns',
                                                         'target_luns')]):
            return zip(connection_properties['target_portals'],
                       connection_properties['target_iqns'],
                       connection_properties['target_luns'])

        return [(connection_properties['target_portal'],
                 connection_properties['target_iqn'],
                 connection_properties.get('target_lun', 0))]
